import functools
import hashlib
import inspect
import os
import sys
import tempfile
import zlib

import matplotlib.image as mpimg
import numpy as np
import pytest
import yaml
from matplotlib.testing.compare import compare_images

import yt.visualization.particle_plots as particle_plots
import yt.visualization.plot_window as pw
import yt.visualization.profile_plotter as profile_plotter
from yt.config import ytcfg
from yt.data_objects.selection_objects.region import YTRegion
from yt.data_objects.static_output import Dataset
from yt.frontends.ytdata.api import save_as_dataset
from yt.loaders import load, load_simulation
from yt.testing import assert_equal
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.on_demand_imports import _h5py as h5py
from yt.visualization.volume_rendering.scene import Scene


def _streamline_for_io(params):
    r"""
    Put test results in a more io-friendly format.

    Many of yt's tests use objects such as tuples as test parameters
    (fields, for instance), but when these objects are written to a
    yaml file, yaml includes python specific anchors that make the file
    harder to read and less portable. The goal of this function is to
    convert these objects to strings (using __repr__() has it's own
    issues) in order to solve this problem.

    Parameters
    ----------
    params : dict
        The dictionary of test parameters in the form
        {param_name : param_value}.

    Returns
    -------
    streamlined_params : dict
        The dictionary of parsed and converted
        {param_name : param_value} pairs.
    """
    streamlined_params = {}
    for key, value in params.items():
        # Check for user-defined functions
        if inspect.isfunction(key):
            key = key.__name__
        if inspect.isfunction(value):
            value = value.__name__
        # The key can be nested iterables, e.g.,
        # d = [None, ('sphere', (center, (0.1, 'unitary')))] so we need
        # to use recursion
        if not isinstance(key, str) and hasattr(key, "__iter__"):
            key = _iterable_to_string(key)
        # The value can also be nested iterables
        if not isinstance(value, str) and hasattr(value, "__iter__"):
            value = _iterable_to_string(value)
        # Scene objects need special treatment to make them more IO friendly
        if isinstance(value, Scene):
            value = "Scene"
        elif isinstance(value, YTRegion):
            value = "Region"
        streamlined_params[key] = value
    return streamlined_params


def _iterable_to_string(iterable):
    r"""
    An extension of streamline_for_io that does the work of making an
    iterable more io-friendly.

    Parameters
    ----------
    iterable : python iterable
        The object to be parsed and converted.

    Returns
    -------
    result : str
        The io-friendly version of the passed iterable.
    """
    result = iterable.__class__.__name__
    for elem in iterable:
        # Check for user-defined functions
        if inspect.isfunction(elem):
            result += "_" + elem.__name__
        # Non-string iterables (e.g., lists, tuples, etc.)
        elif not isinstance(elem, str) and hasattr(elem, "__iter__"):
            result += "_" + _iterable_to_string(elem)
        # Non-string non-iterables (ints, floats, etc.)
        elif not isinstance(elem, str) and not hasattr(elem, "__iter__"):
            result += "_" + str(elem)
        # Strings
        elif isinstance(elem, str):
            result += "_" + elem
    return result


def _hash_results(results):
    r"""
    Driver function for hashing the test result.

    Parameters
    ----------
    results : dict
        Dictionary of {test_name : test_result} pairs.

    Returns
    -------
    results : dict
        Same as the passed results, but the test_results are now
        hex digests of the hashed test_result.
    """
    # Here, results should be comprised of only the tests, not the test
    # parameters
    # Use a new dictionary so as to not overwrite the non-hashed test
    # results in case those are to be saved
    hashed_results = {}
    for test_name, test_value in results.items():
        hashed_results[test_name] = generate_hash(test_value)
    return hashed_results


def _hash_dict(data):
    r"""
    Specifically handles hashing a dictionary object.

    Parameters
    ----------
    data : dict
        The dictionary to be hashed.

    Returns
    -------
    hd.hexdigest : str
        The hex digest of the hashed dictionary.
    """
    hd = None
    for key, value in data.items():
        # Some keys are tuples, not strings
        if not isinstance(key, str):
            key = key.__repr__()
        # Test suites can return values that are dictionaries of other tests
        if isinstance(value, dict):
            hashed_data = _hash_dict(value)
        else:
            hashed_data = bytearray(key.encode("utf8")) + bytearray(value)
        # If we're returning from a recusive call (and therefore hashed_data
        # is a hex digest), we need to encode the string before it can be
        # hashed
        if isinstance(hashed_data, str):
            hashed_data = hashed_data.encode("utf8")
        if hd is None:
            hd = hashlib.md5(hashed_data)
        else:
            hd.update(hashed_data)
    return hd.hexdigest()


def generate_hash(data):
    r"""
    Actually performs the hash operation.

    Parameters
    ----------
    data : python object
        Data to be hashed.

    Returns
    -------
    hd : str
        Hex digest of the hashed data.
    """
    # Sometimes md5 complains that the data is not contiguous, so we
    # make it so here
    if isinstance(data, np.ndarray):
        data = np.ascontiguousarray(data)
    elif isinstance(data, str):
        data = data.encode("utf-8")
    # Try to hash. Some tests return hashable types (like ndarrays) and
    # others don't (such as dictionaries)
    try:
        hd = hashlib.md5(data).hexdigest()
        # Handle those tests that return non-hashable types. This is done
        # here instead of in the tests themselves to try and reduce boilerplate
        # and provide a central location where all of this is done in case it needs
        # to be changed
    except TypeError:
        if isinstance(data, dict):
            hd = _hash_dict(data)
        elif data is None:
            hd = hashlib.md5(bytes(str(-1).encode("utf-8"))).hexdigest()
        else:
            raise
    return hd


def _save_result(data, output_file):
    r"""
    Saves the test results to the desired answer file.

    Parameters
    ----------
    data : dict
        Test results to be saved.

    output_file : str
        Name of file to save results to.
    """
    with open(output_file, "a") as f:
        yaml.dump(data, f)


def _save_raw_arrays(arrays, answer_file, func_name):
    r"""
    Saves the raw arrays produced from answer tests to a file.

    The structure of `answer_file` is: each test function (e.g.,
    test_toro1d[0-None-None-0]) forms a group. Within each group is a
    hdf5 dataset named after the test (e.g., field_values). The value
    stored in each dataset is the raw array corresponding to that
    test and function.

    Parameters
    ----------
    arrays : dict
        Keys are the test name (e.g. field_values) and the values are
        the actual answer arrays produced by the test.

    answer_file : str
        The name of the file to save the answers to, in hdf5 format.

    func_name : str
        The name of the function (possibly augmented by pytest with
        test parameters) that called the test functions
        (e.g, test_toro1d).
    """
    with h5py.File(answer_file, "a") as f:
        grp = f.create_group(func_name)
        for test_name, test_data in arrays.items():
            # Some answer tests (e.g., grid_values, projection_values)
            # return a dictionary, which cannot be handled by h5py
            if isinstance(test_data, dict):
                sub_grp = grp.create_group(test_name)
                _parse_raw_answer_dict(test_data, sub_grp)
            else:
                # Some tests return None, which hdf5 can't handle, and there is
                # no proxy, so we have to make one ourselves. Using -1
                if test_data is None:
                    test_data = -1
                grp.create_dataset(test_name, data=test_data)


def _parse_raw_answer_dict(d, h5grp):
    for k, v in d.items():
        if isinstance(v, dict):
            h5_sub_grp = h5grp.create_group(k)
            _parse_raw_answer_dict(v, h5_sub_grp)
        else:
            k = str(k)
            h5grp.create_dataset(k, data=v)


def _compare_raw_arrays(arrays, answer_file, func_name):
    r"""
    Reads in previously saved raw array data and compares the current
    results with the old ones.

    The structure of `answer_file` is: each test function (e.g.,
    test_toro1d[0-None-None-0]) forms a group. Within each group is a
    hdf5 dataset named after the test (e.g., field_values). The value
    stored in each dataset is the raw array corresponding to that
    test and function.

    Parameters
    ----------
    arrays : dict
        Keys are the test name (e.g. field_values) and the values are
        the actual answer arrays produced by the test.

    answer_file : str
        The name of the file to load the answers from, in hdf5 format.

    func_name : str
        The name of the function (possibly augmented by pytest with
        test parameters) that called the test functions
        (e.g, test_toro1d).
    """
    with h5py.File(answer_file, "r") as f:
        for test_name, new_answer in arrays.items():
            np.testing.assert_array_equal(f[func_name][test_name][:], new_answer)


def can_run_ds(ds_fn, file_check=False):
    r"""
    Validates whether or not a given input can be loaded and used as a
    Dataset object.
    """
    if isinstance(ds_fn, Dataset):
        return True
    path = ytcfg.get("yt", "test_data_dir")
    if not os.path.isdir(path):
        return False
    if file_check:
        return os.path.isfile(os.path.join(path, ds_fn))
    try:
        load(ds_fn)
        return True
    except FileNotFoundError:
        return False


def can_run_sim(sim_fn, sim_type, file_check=False):
    r"""
    Validates whether or not a given input can be used as a simulation
    time-series object.
    """
    path = ytcfg.get("yt", "test_data_dir")
    if not os.path.isdir(path):
        return False
    if file_check:
        return os.path.isfile(os.path.join(path, sim_fn))
    try:
        load_simulation(sim_fn, sim_type)
    except FileNotFoundError:
        return False
    return True


def data_dir_load(ds_fn, cls=None, args=None, kwargs=None):
    r"""
    Loads a sample dataset from the designated test_data_dir for use in
    testing.
    """
    args = args or ()
    kwargs = kwargs or {}
    path = ytcfg.get("yt", "test_data_dir")
    # Some frontends require their field_lists during test parameterization.
    # If the data isn't found, the parameterizing functions return None, since
    # pytest.skip cannot be called outside of a test or fixture.
    if ds_fn is None:
        raise FileNotFoundError
    if not os.path.isdir(path):
        raise FileNotFoundError
    if isinstance(ds_fn, Dataset):
        return ds_fn
    if cls is None:
        ds = load(ds_fn, *args, **kwargs)
    else:
        ds = cls(os.path.join(path, ds_fn), *args, **kwargs)
    ds.index
    return ds


def requires_ds(ds_fn, file_check=False):
    r"""
    Meta-wrapper for specifying required data for a test and
    checking if said data exists.
    """

    def ffalse(func):
        @functools.wraps(func)
        def skip(*args, **kwargs):
            msg = f"{ds_fn} not found, skipping {func.__name__}."
            pytest.skip(msg)

        return skip

    def ftrue(func):
        @functools.wraps(func)
        def true_wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return true_wrapper

    if not can_run_ds(ds_fn, file_check):
        return ffalse
    else:
        return ftrue


def requires_sim(sim_fn, sim_type, file_check=False):
    r"""
    Meta-wrapper for specifying a required simulation for a test and
    checking if said simulation exists.
    """

    def ffalse(func):
        @functools.wraps(func)
        def skip(*args, **kwargs):
            msg = f"{sim_fn} not found, skipping {func.__name__}."
            pytest.skip(msg)

        return skip

    def ftrue(func):
        @functools.wraps(func)
        def true_wrapper(*args, **kwargs):
            return func

        return true_wrapper

    if not can_run_sim(sim_fn, sim_type, file_check):
        return ffalse
    else:
        return ftrue


def create_obj(ds, obj_type):
    # obj_type should be tuple of
    #  ( obj_name, ( args ) )
    if obj_type is None:
        return ds.all_data()
    cls = getattr(ds, obj_type[0])
    obj = cls(*obj_type[1])
    return obj


def compare_unit_attributes(ds1, ds2):
    r"""
    Checks to make sure that the length, mass, time, velocity, and
    magnetic units are the same for two different dataset objects.
    """
    attrs = ("length_unit", "mass_unit", "time_unit", "velocity_unit", "magnetic_unit")
    for attr in attrs:
        u1 = getattr(ds1, attr, None)
        u2 = getattr(ds2, attr, None)
        assert u1 == u2


def fake_halo_catalog(data):
    filename = "catalog.0.h5"
    ftypes = {field: "." for field in data}
    extra_attrs = {"data_type": "halo_catalog", "num_halos": data["particle_mass"].size}
    ds = {
        "cosmological_simulation": 1,
        "omega_lambda": 0.7,
        "omega_matter": 0.3,
        "hubble_constant": 0.7,
        "current_redshift": 0,
        "current_time": YTQuantity(1, "yr"),
        "domain_left_edge": YTArray(np.zeros(3), "cm"),
        "domain_right_edge": YTArray(np.ones(3), "cm"),
    }
    save_as_dataset(ds, filename, data, field_types=ftypes, extra_attrs=extra_attrs)
    return filename


def _create_plot_window_attribute_plot(ds, plot_type, field, axis, pkwargs=None):
    r"""
    Convenience function used in plot_window_attribute_test.

    Parameters
    ----------
    ds : Dataset
        The Dataset object from which the plotting data is taken.

    plot_type : string
        Type of plot to make (e.g., SlicePlot).

    field : yt field
        The field (e.g, density) to plot.

    axis : int
        The plot axis to plot or project along.

    pkwargs : dict
        Any keywords to be passed when creating the plot.
    """
    if plot_type is None:
        raise RuntimeError("Must explicitly request a plot type")
    cls = getattr(pw, plot_type, None)
    if cls is None:
        cls = getattr(particle_plots, plot_type)
    plot = cls(ds, axis, field, **pkwargs)
    return plot


def _create_phase_plot_attribute_plot(
    data_source, x_field, y_field, z_field, plot_type, plot_kwargs=None
):
    r"""
    Convenience function used in phase_plot_attribute_test.

    Parameters
    ----------
    data_source : Dataset object
        The Dataset object from which the plotting data is taken.

    x_field : yt field
        Field to plot on x-axis.

    y_field : yt field
        Field to plot on y-axis.

    z_field : yt field
        Field to plot on z-axis.

    plot_type : string
        Type of plot to make (e.g., SlicePlot).

    plot_kwargs : dict
        Any keywords to be passed when creating the plot.
    """
    if plot_type is None:
        raise RuntimeError("Must explicitly request a plot type")
    cls = getattr(profile_plotter, plot_type, None)
    if cls is None:
        cls = getattr(particle_plots, plot_type)
    plot = cls(*(data_source, x_field, y_field, z_field), **plot_kwargs)
    return plot


def get_parameterization(fname):
    """
    Returns a dataset's field list to make test parameterizationn easier.

    Some tests (such as those that use the toro1d dataset in enzo) check
    every field in a dataset. In order to parametrize the tests without
    having to hardcode a list of every field, this function is used.
    Additionally, if the dataset cannot be found, this function enables
    pytest to mark the test as failed without the whole test run crashing,
    since the parameterization happens at import time.
    """
    try:
        ds = data_dir_load(fname)
        return ds.field_list
    except FileNotFoundError:
        return [
            None,
        ]


def compare_image_lists(new_result, old_result, decimals):
    fns = []
    for _i in range(2):
        tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
        os.close(tmpfd)
        fns.append(tmpname)
    num_images = len(old_result)
    assert num_images > 0
    for i in range(num_images):
        mpimg.imsave(fns[0], np.loads(zlib.decompress(old_result[i])))
        mpimg.imsave(fns[1], np.loads(zlib.decompress(new_result[i])))
        results = compare_images(fns[0], fns[1], 10 ** (-decimals))
        if results is not None:
            if os.environ.get("JENKINS_HOME") is not None:
                tempfiles = [
                    line.strip()
                    for line in results.split("\n")
                    if line.endswith(".png")
                ]
                for fn in tempfiles:
                    sys.stderr.write(f"\n[[ATTACHMENT|{fn}]]")
                sys.stderr.write("\n")
        assert_equal(results, None, results)
        for fn in fns:
            os.remove(fn)



import collections
import errno
import os
import struct
import weakref

import numpy as np

from yt.data_objects.index_subobjects.particle_container import ParticleContainer
from yt.funcs import get_pbar, only_on_root
from yt.geometry.geometry_handler import Index, YTDataChunk
from yt.geometry.particle_oct_container import ParticleBitmap
from yt.utilities.lib.fnv_hash import fnv_hash
from yt.utilities.logger import ytLogger as mylog


class ParticleIndex(Index):
    """The Index subclass for particle datasets"""

    def __init__(self, ds, dataset_type):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        self.float_type = np.float64
        super().__init__(ds, dataset_type)
        self._initialize_index()

    def _setup_geometry(self):
        self.regions = None

    def get_smallest_dx(self):
        """
        Returns (in code units) the smallest cell size in the simulation.
        """
        return self.ds.arr(0, "code_length")

    def _get_particle_type_counts(self):
        result = collections.defaultdict(lambda: 0)
        for df in self.data_files:
            for k in df.total_particles.keys():
                result[k] += df.total_particles[k]
        return dict(result)

    def convert(self, unit):
        return self.dataset.conversion_factors[unit]

    @property
    def chunksize(self):
        # This can be overridden in subclasses
        return 64 ** 3

    _data_files = None

    @property
    def data_files(self):
        if self._data_files is not None:
            return self._data_files

        self._setup_filenames()
        return self._data_files

    @data_files.setter
    def data_files(self, value):
        self._data_files = value

    _total_particles = None

    @property
    def total_particles(self):
        if self._total_particles is not None:
            return self._total_particles

        self._total_particles = sum(
            sum(d.total_particles.values()) for d in self.data_files
        )
        return self._total_particles

    def _setup_filenames(self):
        template = self.dataset.filename_template
        ndoms = self.dataset.file_count
        cls = self.dataset._file_class
        self.data_files = []
        fi = 0
        for i in range(int(ndoms)):
            start = 0
            if self.chunksize > 0:
                end = start + self.chunksize
            else:
                end = None
            while True:
                df = cls(self.dataset, self.io, template % {"num": i}, fi, (start, end))
                if max(df.total_particles.values()) == 0:
                    break
                fi += 1
                self.data_files.append(df)
                if self.chunksize <= 0:
                    break
                start = end
                end += self.chunksize

    def _initialize_index(self):
        ds = self.dataset
        only_on_root(
            mylog.info,
            "Allocating for %0.3e particles",
            self.total_particles,
            global_rootonly=True,
        )

        # if we have not yet set domain_left_edge and domain_right_edge then do
        # an I/O pass over the particle coordinates to determine a bounding box
        if self.ds.domain_left_edge is None:
            min_ppos = np.empty(3, dtype="float64")
            min_ppos[:] = np.nan
            max_ppos = np.empty(3, dtype="float64")
            max_ppos[:] = np.nan
            only_on_root(
                mylog.info,
                "Bounding box cannot be inferred from metadata, reading "
                "particle positions to infer bounding box",
            )
            for df in self.data_files:
                for _, ppos in self.io._yield_coordinates(df):
                    min_ppos = np.nanmin(np.vstack([min_ppos, ppos]), axis=0)
                    max_ppos = np.nanmax(np.vstack([max_ppos, ppos]), axis=0)
            only_on_root(
                mylog.info,
                "Load this dataset with bounding_box=[%s, %s] to avoid I/O "
                "overhead from inferring bounding_box." % (min_ppos, max_ppos),
            )
            ds.domain_left_edge = ds.arr(1.05 * min_ppos, "code_length")
            ds.domain_right_edge = ds.arr(1.05 * max_ppos, "code_length")
            ds.domain_width = ds.domain_right_edge - ds.domain_left_edge

        # use a trivial morton index for datasets containing a single chunk
        if len(self.data_files) == 1:
            order1 = 1
            order2 = 1
        else:
            order1 = ds.index_order[0]
            order2 = ds.index_order[1]

        if order1 == 1 and order2 == 1:
            dont_cache = True
        else:
            dont_cache = False

        # If we have applied a bounding box then we can't cache the
        # ParticleBitmap because it is doman dependent
        if getattr(ds, "_domain_override", False):
            dont_cache = True

        if not hasattr(self.ds, "_file_hash"):
            self.ds._file_hash = self._generate_hash()

        self.regions = ParticleBitmap(
            ds.domain_left_edge,
            ds.domain_right_edge,
            ds.periodicity,
            self.ds._file_hash,
            len(self.data_files),
            index_order1=order1,
            index_order2=order2,
        )

        # Load Morton index from file if provided
        def _current_fname():
            if getattr(ds, "index_filename", None) is None:
                fname = ds.parameter_filename + ".index{}_{}.ewah".format(
                    self.regions.index_order1, self.regions.index_order2
                )
            else:
                fname = ds.index_filename
            return fname

        fname = _current_fname()

        dont_load = dont_cache and not hasattr(ds, "index_filename")
        try:
            if dont_load:
                raise OSError
            rflag = self.regions.load_bitmasks(fname)
            rflag = self.regions.check_bitmasks()
            self._initialize_frontend_specific()
            if rflag == 0:
                raise OSError
        except (OSError, struct.error):
            self.regions.reset_bitmasks()
            self._initialize_coarse_index()
            self._initialize_refined_index()
            # We now update fname since index_order2 may have changed
            fname = _current_fname()
            wdir = os.path.dirname(fname)
            if not dont_cache and os.access(wdir, os.W_OK):
                # Sometimes os mis-reports whether a directory is writable,
                # So pass if writing the bitmask file fails.
                try:
                    self.regions.save_bitmasks(fname)
                except OSError:
                    pass
            rflag = self.regions.check_bitmasks()

    def _initialize_coarse_index(self):
        pb = get_pbar("Initializing coarse index ", len(self.data_files))
        max_hsml = 0.0
        for i, data_file in enumerate(self.data_files):
            pb.update(i + 1)
            for ptype, pos in self.io._yield_coordinates(data_file):
                ds = self.ds
                if hasattr(ds, "_sph_ptypes") and ptype == ds._sph_ptypes[0]:
                    hsml = self.io._get_smoothing_length(
                        data_file, pos.dtype, pos.shape
                    )
                    if hsml is not None and hsml.size > 0.0:
                        max_hsml = max(max_hsml, hsml.max())
                else:
                    hsml = None
                self.regions._coarse_index_data_file(pos, hsml, data_file.file_id)
            self.regions._set_coarse_index_data_file(data_file.file_id)
        pb.finish()
        self.regions.find_collisions_coarse()
        if max_hsml > 0.0 and len(self.data_files) > 1:
            # By passing this in, we only allow index_order2 to be increased by
            # two at most, never increased.  One place this becomes particularly
            # useful is in the case of an extremely small section of gas
            # particles embedded in a much much larger domain.  The max
            # smoothing length will be quite small, so based on the larger
            # domain, it will correspond to a very very high index order, which
            # is a large amount of memory!  Having multiple indexes, one for
            # each particle type, would fix this.
            new_order2 = self.regions.update_mi2(max_hsml, ds.index_order[1] + 2)
            mylog.info(
                "Updating index_order2 from %s to %s", ds.index_order[1], new_order2
            )
            self.ds.index_order = (self.ds.index_order[0], new_order2)

    def _initialize_refined_index(self):
        mask = self.regions.masks.sum(axis=1).astype("uint8")
        max_npart = max(sum(d.total_particles.values()) for d in self.data_files) * 28
        sub_mi1 = np.zeros(max_npart, "uint64")
        sub_mi2 = np.zeros(max_npart, "uint64")
        pb = get_pbar("Initializing refined index", len(self.data_files))
        mask_threshold = getattr(self, "_index_mask_threshold", 2)
        count_threshold = getattr(self, "_index_count_threshold", 256)
        mylog.debug(
            "Using estimated thresholds of %s and %s for refinement",
            mask_threshold,
            count_threshold,
        )
        total_refined = 0
        total_coarse_refined = (
            (mask >= 2) & (self.regions.particle_counts > count_threshold)
        ).sum()
        mylog.debug(
            "This should produce roughly %s zones, for %s of the domain",
            total_coarse_refined,
            100 * total_coarse_refined / mask.size,
        )
        for i, data_file in enumerate(self.data_files):
            coll = None
            pb.update(i + 1)
            nsub_mi = 0
            for ptype, pos in self.io._yield_coordinates(data_file):
                if pos.size == 0:
                    continue
                if hasattr(self.ds, "_sph_ptypes") and ptype == self.ds._sph_ptypes[0]:
                    hsml = self.io._get_smoothing_length(
                        data_file, pos.dtype, pos.shape
                    )
                else:
                    hsml = None
                nsub_mi, coll = self.regions._refined_index_data_file(
                    coll,
                    pos,
                    hsml,
                    mask,
                    sub_mi1,
                    sub_mi2,
                    data_file.file_id,
                    nsub_mi,
                    count_threshold=count_threshold,
                    mask_threshold=mask_threshold,
                )
                total_refined += nsub_mi
            self.regions.bitmasks.append(data_file.file_id, coll)
        pb.finish()
        self.regions.find_collisions_refined()

    def _detect_output_fields(self):
        # TODO: Add additional fields
        dsl = []
        units = {}
        pcounts = self._get_particle_type_counts()
        field_cache = {}
        for dom in self.data_files:
            if dom.filename in field_cache:
                fl, _units = field_cache[dom.filename]
            else:
                fl, _units = self.io._identify_fields(dom)
                field_cache[dom.filename] = fl, _units
            units.update(_units)
            dom._calculate_offsets(fl, pcounts)
            for f in fl:
                if f not in dsl:
                    dsl.append(f)
        self.field_list = dsl
        ds = self.dataset
        ds.particle_types = tuple({pt for pt, ds in dsl})
        # This is an attribute that means these particle types *actually*
        # exist.  As in, they are real, in the dataset.
        ds.field_units.update(units)
        ds.particle_types_raw = ds.particle_types

    def _identify_base_chunk(self, dobj):
        # Must check that chunk_info contains the right number of ghost zones
        if getattr(dobj, "_chunk_info", None) is None:
            if isinstance(dobj, ParticleContainer):
                dobj._chunk_info = [dobj]
            else:
                # TODO: only return files
                if getattr(dobj.selector, "is_all_data", False):
                    nfiles = self.regions.nfiles
                    dfi = np.arange(nfiles)
                else:
                    dfi, file_masks, addfi = self.regions.identify_file_masks(
                        dobj.selector
                    )
                    nfiles = len(file_masks)
                dobj._chunk_info = [None for _ in range(nfiles)]

                # The following was moved here from ParticleContainer in order
                # to make the ParticleContainer object pickleable. By having
                # the base_selector as its own argument, we avoid having to
                # rebuild the index on unpickling a ParticleContainer.
                if hasattr(dobj, "base_selector"):
                    base_selector = dobj.base_selector
                    base_region = dobj.base_region
                else:
                    base_region = dobj
                    base_selector = dobj.selector

                for i in range(nfiles):
                    domain_id = i + 1
                    dobj._chunk_info[i] = ParticleContainer(
                        base_region,
                        base_selector,
                        [self.data_files[dfi[i]]],
                        domain_id=domain_id,
                    )
                # NOTE: One fun thing about the way IO works is that it
                # consolidates things quite nicely.  So we should feel free to
                # create as many objects as part of the chunk as we want, since
                # it'll take the set() of them.  So if we break stuff up like
                # this here, we end up in a situation where we have the ability
                # to break things down further later on for buffer zones and the
                # like.
        (dobj._current_chunk,) = self._chunk_all(dobj)

    def _chunk_all(self, dobj):
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        yield YTDataChunk(dobj, "all", oobjs, None)

    def _chunk_spatial(self, dobj, ngz, sort=None, preload_fields=None):
        sobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for og in sobjs:
            with og._expand_data_files():
                if ngz > 0:
                    g = og.retrieve_ghost_zones(ngz, [], smoothed=True)
                else:
                    g = og
                yield YTDataChunk(dobj, "spatial", [g])

    def _chunk_io(self, dobj, cache=True, local_only=False):
        oobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        for container in oobjs:
            yield YTDataChunk(dobj, "io", [container], None, cache=cache)

    def _generate_hash(self):
        # Generate an FNV hash by creating a byte array containing the
        # modification time of as well as the first and last 1 MB of data in
        # every output file
        ret = bytearray()
        for pfile in self.data_files:

            # only look at "real" files, not "fake" files generated by the
            # chunking system
            if pfile.start not in (0, None):
                continue
            try:
                mtime = os.path.getmtime(pfile.filename)
            except OSError as e:
                if e.errno == errno.ENOENT:
                    # this is an in-memory file so we return with a dummy
                    # value
                    return -1
                else:
                    raise
            ret.extend(str(mtime).encode("utf-8"))
            size = os.path.getsize(pfile.filename)
            if size > 1e6:
                size = int(1e6)
            with open(pfile.filename, "rb") as fh:
                # read in first and last 1 MB of data
                data = fh.read(size)
                fh.seek(-size, os.SEEK_END)
                data = fh.read(size)
                ret.extend(data)
            return fnv_hash(ret)

    def _initialize_frontend_specific(self):
        """This is for frontend-specific initialization code

        If there are frontend-specific things that need to be set while
        creating the index, this function forces these operations to happen
        in cases where we are reloading the index from a sidecar file.
        """
        pass



import abc
import weakref
from collections import defaultdict

import numpy as np

from yt.arraytypes import blankRecordArray
from yt.config import ytcfg
from yt.fields.derived_field import ValidateSpatial
from yt.fields.field_detector import FieldDetector
from yt.funcs import ensure_numpy_array, iter_fields
from yt.geometry.geometry_handler import ChunkDataCache, Index, YTDataChunk
from yt.utilities.definitions import MAXLEVEL
from yt.utilities.logger import ytLogger as mylog

from .grid_container import GridTree, MatchPointsToGrids


class GridIndex(Index, abc.ABC):
    """The index class for patch and block AMR datasets."""

    float_type = "float64"
    _preload_implemented = False
    _index_properties = (
        "grid_left_edge",
        "grid_right_edge",
        "grid_levels",
        "grid_particle_count",
        "grid_dimensions",
    )

    def _setup_geometry(self):
        mylog.debug("Counting grids.")
        self._count_grids()

        mylog.debug("Initializing grid arrays.")
        self._initialize_grid_arrays()

        mylog.debug("Parsing index.")
        self._parse_index()

        mylog.debug("Constructing grid objects.")
        self._populate_grid_objects()

        mylog.debug("Re-examining index")
        self._initialize_level_stats()

    @abc.abstractmethod
    def _count_grids(self):
        pass

    @abc.abstractmethod
    def _parse_index(self):
        pass

    @abc.abstractmethod
    def _populate_grid_objects(self):
        pass

    def __del__(self):
        del self.grid_dimensions
        del self.grid_left_edge
        del self.grid_right_edge
        del self.grid_levels
        del self.grid_particle_count
        del self.grids

    @property
    def parameters(self):
        return self.dataset.parameters

    def _detect_output_fields_backup(self):
        # grab fields from backup file as well, if present
        return

    def select_grids(self, level):
        """
        Returns an array of grids at *level*.
        """
        return self.grids[self.grid_levels.flat == level]

    def get_levels(self):
        for level in range(self.max_level + 1):
            yield self.select_grids(level)

    def _initialize_grid_arrays(self):
        mylog.debug("Allocating arrays for %s grids", self.num_grids)
        self.grid_dimensions = np.ones((self.num_grids, 3), "int32")
        self.grid_left_edge = self.ds.arr(
            np.zeros((self.num_grids, 3), self.float_type), "code_length"
        )
        self.grid_right_edge = self.ds.arr(
            np.ones((self.num_grids, 3), self.float_type), "code_length"
        )
        self.grid_levels = np.zeros((self.num_grids, 1), "int32")
        self.grid_particle_count = np.zeros((self.num_grids, 1), "int32")

    def clear_all_data(self):
        """
        This routine clears all the data currently being held onto by the grids
        and the data io handler.
        """
        for g in self.grids:
            g.clear_data()
        self.io.queue.clear()

    def get_smallest_dx(self):
        """
        Returns (in code units) the smallest cell size in the simulation.
        """
        return self.select_grids(self.grid_levels.max())[0].dds[:].min()

    def _get_particle_type_counts(self):
        return {self.ds.particle_types_raw[0]: self.grid_particle_count.sum()}

    def _initialize_level_stats(self):
        # Now some statistics:
        #   0 = number of grids
        #   1 = number of cells
        #   2 = blank
        desc = {"names": ["numgrids", "numcells", "level"], "formats": ["int64"] * 3}
        self.level_stats = blankRecordArray(desc, MAXLEVEL)
        self.level_stats["level"] = [i for i in range(MAXLEVEL)]
        self.level_stats["numgrids"] = [0 for i in range(MAXLEVEL)]
        self.level_stats["numcells"] = [0 for i in range(MAXLEVEL)]
        for level in range(self.max_level + 1):
            self.level_stats[level]["numgrids"] = np.sum(self.grid_levels == level)
            li = self.grid_levels[:, 0] == level
            self.level_stats[level]["numcells"] = (
                self.grid_dimensions[li, :].prod(axis=1).sum()
            )

    @property
    def grid_corners(self):
        return np.array(
            [
                [
                    self.grid_left_edge[:, 0],
                    self.grid_left_edge[:, 1],
                    self.grid_left_edge[:, 2],
                ],
                [
                    self.grid_right_edge[:, 0],
                    self.grid_left_edge[:, 1],
                    self.grid_left_edge[:, 2],
                ],
                [
                    self.grid_right_edge[:, 0],
                    self.grid_right_edge[:, 1],
                    self.grid_left_edge[:, 2],
                ],
                [
                    self.grid_left_edge[:, 0],
                    self.grid_right_edge[:, 1],
                    self.grid_left_edge[:, 2],
                ],
                [
                    self.grid_left_edge[:, 0],
                    self.grid_left_edge[:, 1],
                    self.grid_right_edge[:, 2],
                ],
                [
                    self.grid_right_edge[:, 0],
                    self.grid_left_edge[:, 1],
                    self.grid_right_edge[:, 2],
                ],
                [
                    self.grid_right_edge[:, 0],
                    self.grid_right_edge[:, 1],
                    self.grid_right_edge[:, 2],
                ],
                [
                    self.grid_left_edge[:, 0],
                    self.grid_right_edge[:, 1],
                    self.grid_right_edge[:, 2],
                ],
            ],
            dtype="float64",
        )

    def lock_grids_to_parents(self):
        r"""This function locks grid edges to their parents.

        This is useful in cases where the grid structure may be somewhat
        irregular, or where setting the left and right edges is a lossy
        process.  It is designed to correct situations where left/right edges
        may be set slightly incorrectly, resulting in discontinuities in images
        and the like.
        """
        mylog.info("Locking grids to parents.")
        for i, g in enumerate(self.grids):
            si = g.get_global_startindex()
            g.LeftEdge = self.ds.domain_left_edge + g.dds * si
            g.RightEdge = g.LeftEdge + g.ActiveDimensions * g.dds
            self.grid_left_edge[i, :] = g.LeftEdge
            self.grid_right_edge[i, :] = g.RightEdge

    def print_stats(self):
        """
        Prints out (stdout) relevant information about the simulation
        """
        header = "{:>3}\t{:>6}\t{:>14}\t{:>14}".format(
            "level", "# grids", "# cells", "# cells^3"
        )
        print(header)
        print(f"{len(header.expandtabs()) * '-'}")
        for level in range(MAXLEVEL):
            if (self.level_stats["numgrids"][level]) == 0:
                continue
            print(
                "% 3i\t% 6i\t% 14i\t% 14i"
                % (
                    level,
                    self.level_stats["numgrids"][level],
                    self.level_stats["numcells"][level],
                    np.ceil(self.level_stats["numcells"][level] ** (1.0 / 3)),
                )
            )
            dx = self.select_grids(level)[0].dds[0]
        print("-" * 46)
        print(
            "   \t% 6i\t% 14i"
            % (self.level_stats["numgrids"].sum(), self.level_stats["numcells"].sum())
        )
        print("\n")
        try:
            print(f"z = {self['CosmologyCurrentRedshift']:0.8f}")
        except Exception:
            pass
        print(
            "t = %0.8e = %0.8e s = %0.8e years"
            % (
                self.ds.current_time.in_units("code_time"),
                self.ds.current_time.in_units("s"),
                self.ds.current_time.in_units("yr"),
            )
        )
        print("\nSmallest Cell:")
        for item in ("Mpc", "pc", "AU", "cm"):
            print(f"\tWidth: {dx.in_units(item):0.3e} {item}")

    def _find_field_values_at_points(self, fields, coords):
        r"""Find the value of fields at a set of coordinates.

        Returns the values [field1, field2,...] of the fields at the given
        (x, y, z) points. Returns a numpy array of field values cross coords
        """
        coords = self.ds.arr(ensure_numpy_array(coords), "code_length")
        grids = self._find_points(coords[:, 0], coords[:, 1], coords[:, 2])[0]
        fields = list(iter_fields(fields))
        mark = np.zeros(3, dtype="int64")
        out = []

        # create point -> grid mapping
        grid_index = {}
        for coord_index, grid in enumerate(grids):
            if grid not in grid_index:
                grid_index[grid] = []
            grid_index[grid].append(coord_index)

        out = []
        for field in fields:
            funit = self.ds._get_field_info(field).units
            out.append(self.ds.arr(np.empty(len(coords)), funit))

        for grid in grid_index:
            cellwidth = (grid.RightEdge - grid.LeftEdge) / grid.ActiveDimensions
            for field_index, field in enumerate(fields):
                for coord_index in grid_index[grid]:
                    mark = (coords[coord_index, :] - grid.LeftEdge) / cellwidth
                    mark = np.array(mark, dtype="int64")
                    out[field_index][coord_index] = grid[field][
                        mark[0], mark[1], mark[2]
                    ]
        if len(fields) == 1:
            return out[0]
        return out

    def _find_points(self, x, y, z):
        """
        Returns the (objects, indices) of leaf grids
        containing a number of (x,y,z) points
        """
        x = ensure_numpy_array(x)
        y = ensure_numpy_array(y)
        z = ensure_numpy_array(z)
        if not len(x) == len(y) == len(z):
            raise ValueError("Arrays of indices must be of the same size")

        grid_tree = self._get_grid_tree()
        pts = MatchPointsToGrids(grid_tree, len(x), x, y, z)
        ind = pts.find_points_in_tree()
        return self.grids[ind], ind

    def _get_grid_tree(self):

        left_edge = self.ds.arr(np.zeros((self.num_grids, 3)), "code_length")
        right_edge = self.ds.arr(np.zeros((self.num_grids, 3)), "code_length")
        level = np.zeros((self.num_grids), dtype="int64")
        parent_ind = np.zeros((self.num_grids), dtype="int64")
        num_children = np.zeros((self.num_grids), dtype="int64")
        dimensions = np.zeros((self.num_grids, 3), dtype="int32")

        for i, grid in enumerate(self.grids):

            left_edge[i, :] = grid.LeftEdge
            right_edge[i, :] = grid.RightEdge
            level[i] = grid.Level
            if grid.Parent is None:
                parent_ind[i] = -1
            else:
                parent_ind[i] = grid.Parent.id - grid.Parent._id_offset
            num_children[i] = np.int64(len(grid.Children))
            dimensions[i, :] = grid.ActiveDimensions

        return GridTree(
            self.num_grids,
            left_edge,
            right_edge,
            dimensions,
            parent_ind,
            level,
            num_children,
        )

    def convert(self, unit):
        return self.dataset.conversion_factors[unit]

    def _identify_base_chunk(self, dobj):
        fast_index = None
        if dobj._type_name == "grid":
            dobj._chunk_info = np.empty(1, dtype="object")
            dobj._chunk_info[0] = weakref.proxy(dobj)
        elif getattr(dobj, "_grids", None) is None:
            gi = dobj.selector.select_grids(
                self.grid_left_edge, self.grid_right_edge, self.grid_levels
            )
            if any([g.filename is not None for g in self.grids[gi]]):
                _gsort = _grid_sort_mixed
            else:
                _gsort = _grid_sort_id
            grids = list(sorted(self.grids[gi], key=_gsort))
            dobj._chunk_info = np.empty(len(grids), dtype="object")
            for i, g in enumerate(grids):
                dobj._chunk_info[i] = g
        # These next two lines, when uncommented, turn "on" the fast index.
        # if dobj._type_name != "grid":
        #    fast_index = self._get_grid_tree()
        if getattr(dobj, "size", None) is None:
            dobj.size = self._count_selection(dobj, fast_index=fast_index)
        if getattr(dobj, "shape", None) is None:
            dobj.shape = (dobj.size,)
        dobj._current_chunk = list(
            self._chunk_all(dobj, cache=False, fast_index=fast_index)
        )[0]

    def _count_selection(self, dobj, grids=None, fast_index=None):
        if fast_index is not None:
            return fast_index.count(dobj.selector)
        if grids is None:
            grids = dobj._chunk_info
        count = sum(g.count(dobj.selector) for g in grids)
        return count

    def _chunk_all(self, dobj, cache=True, fast_index=None):
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        fast_index = fast_index or getattr(dobj._current_chunk, "_fast_index", None)
        yield YTDataChunk(dobj, "all", gobjs, dobj.size, cache, fast_index=fast_index)

    def _chunk_spatial(self, dobj, ngz, sort=None, preload_fields=None):
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        if sort in ("+level", "level"):
            giter = sorted(gobjs, key=lambda g: g.Level)
        elif sort == "-level":
            giter = sorted(gobjs, key=lambda g: -g.Level)
        elif sort is None:
            giter = gobjs
        if preload_fields is None:
            preload_fields = []
        preload_fields, _ = self._split_fields(preload_fields)
        if self._preload_implemented and len(preload_fields) > 0 and ngz == 0:
            giter = ChunkDataCache(list(giter), preload_fields, self)
        for og in giter:
            if ngz > 0:
                g = og.retrieve_ghost_zones(ngz, [], smoothed=True)
            else:
                g = og
            size = self._count_selection(dobj, [og])
            if size == 0:
                continue
            # We don't want to cache any of the masks or icoords or fcoords for
            # individual grids.
            yield YTDataChunk(dobj, "spatial", [g], size, cache=False)

    _grid_chunksize = 1000

    def _chunk_io(
        self,
        dobj,
        cache=True,
        local_only=False,
        preload_fields=None,
        chunk_sizing="auto",
    ):
        # local_only is only useful for inline datasets and requires
        # implementation by subclasses.
        if preload_fields is None:
            preload_fields = []
        preload_fields, _ = self._split_fields(preload_fields)
        gfiles = defaultdict(list)
        gobjs = getattr(dobj._current_chunk, "objs", dobj._chunk_info)
        fast_index = dobj._current_chunk._fast_index
        for g in gobjs:
            # Force to be a string because sometimes g.filename is None.
            gfiles[str(g.filename)].append(g)
        # We can apply a heuristic here to make sure we aren't loading too
        # many grids all at once.
        if chunk_sizing == "auto":
            chunk_ngrids = len(gobjs)
            if chunk_ngrids > 0:
                nproc = int(ytcfg.get("yt", "internals", "global_parallel_size"))
                chunking_factor = np.ceil(
                    self._grid_chunksize * nproc / chunk_ngrids
                ).astype("int")
                size = max(self._grid_chunksize // chunking_factor, 1)
            else:
                size = self._grid_chunksize
        elif chunk_sizing == "config_file":
            size = ytcfg.get("yt", "chunk_size")
        elif chunk_sizing == "just_one":
            size = 1
        elif chunk_sizing == "old":
            size = self._grid_chunksize
        else:
            raise RuntimeError(
                f"{chunk_sizing} is an invalid value for the 'chunk_sizing' argument."
            )
        for fn in sorted(gfiles):
            gs = gfiles[fn]
            for grids in (gs[pos : pos + size] for pos in range(0, len(gs), size)):
                dc = YTDataChunk(
                    dobj,
                    "io",
                    grids,
                    self._count_selection(dobj, grids),
                    cache=cache,
                    fast_index=fast_index,
                )
                # We allow four full chunks to be included.
                with self.io.preload(dc, preload_fields, 4.0 * size):
                    yield dc

    def _add_mesh_sampling_particle_field(self, deposit_field, ftype, ptype):
        units = self.ds.field_info[ftype, deposit_field].units
        take_log = self.ds.field_info[ftype, deposit_field].take_log
        field_name = f"cell_{ftype}_{deposit_field}"

        def _mesh_sampling_particle_field(field, data):
            pos = data[ptype, "particle_position"]
            field_values = data[ftype, deposit_field]

            if isinstance(data, FieldDetector):
                return np.zeros(pos.shape[0])

            i, j, k = np.floor((pos - data.LeftEdge) / data.dds).astype("int64").T

            # Make sure all particles are within the current grid, otherwise return nan
            maxi, maxj, maxk = field_values.shape

            mask = (i < maxi) & (j < maxj) & (k < maxk)
            mask &= (i >= 0) & (j >= 0) & (k >= 0)

            result = np.full(len(pos), np.nan, dtype="float64")
            if result.shape[0] > 0:
                result[mask] = field_values[i[mask], j[mask], k[mask]]

            return data.ds.arr(result, field_values.units)

        self.ds.add_field(
            (ptype, field_name),
            function=_mesh_sampling_particle_field,
            sampling_type="particle",
            units=units,
            take_log=take_log,
            validators=[ValidateSpatial()],
        )


def _grid_sort_id(g):
    return g.id


def _grid_sort_mixed(g):
    if g.filename is None:
        return str(g.id)
    return g.filename



import os
import shutil
import tempfile
import unittest
from unittest import mock

import numpy as np

from yt.data_objects.particle_filters import add_particle_filter
from yt.data_objects.profiles import create_profile
from yt.loaders import load
from yt.testing import (
    assert_allclose,
    assert_array_almost_equal,
    fake_particle_ds,
    requires_file,
)
from yt.units.yt_array import YTArray
from yt.utilities.answer_testing.framework import (
    PhasePlotAttributeTest,
    PlotWindowAttributeTest,
    data_dir_load,
    requires_ds,
)
from yt.visualization.api import ParticlePhasePlot, ParticlePlot, ParticleProjectionPlot
from yt.visualization.tests.test_plotwindow import ATTR_ARGS, WIDTH_SPECS


def setup():
    """Test specific setup."""
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


#  override some of the plotwindow ATTR_ARGS
PROJ_ATTR_ARGS = ATTR_ARGS.copy()
PROJ_ATTR_ARGS["set_cmap"] = [
    ((("all", "particle_mass"), "RdBu"), {}),
    ((("all", "particle_mass"), "kamae"), {}),
]
PROJ_ATTR_ARGS["set_log"] = [((("all", "particle_mass"), False), {})]
PROJ_ATTR_ARGS["set_zlim"] = [
    ((("all", "particle_mass"), 1e39, 1e42), {}),
    ((("all", "particle_mass"), 1e39, None), {"dynamic_range": 4}),
]

PHASE_ATTR_ARGS = {
    "annotate_text": [
        (((5e-29, 5e7), "Hello YT"), {}),
        (((5e-29, 5e7), "Hello YT"), {"color": "b"}),
    ],
    "set_title": [((("all", "particle_mass"), "A phase plot."), {})],
    "set_log": [((("all", "particle_mass"), False), {})],
    "set_unit": [((("all", "particle_mass"), "Msun"), {})],
    "set_xlim": [((-4e7, 4e7), {})],
    "set_ylim": [((-4e7, 4e7), {})],
}

TEST_FLNMS = [None, "test", "test.png", "test.eps", "test.ps", "test.pdf"]

CENTER_SPECS = (
    "c",
    "C",
    "center",
    "Center",
    [0.5, 0.5, 0.5],
    [[0.2, 0.3, 0.4], "cm"],
    YTArray([0.3, 0.4, 0.7], "cm"),
)

WEIGHT_FIELDS = (None, ("all", "particle_ones"), ("all", "particle_mass"))

PHASE_FIELDS = [
    (
        ("all", "particle_velocity_x"),
        ("all", "particle_position_z"),
        ("all", "particle_mass"),
    ),
    (
        ("all", "particle_position_x"),
        ("all", "particle_position_y"),
        ("all", "particle_ones"),
    ),
    (
        ("all", "particle_velocity_x"),
        ("all", "particle_velocity_y"),
        [("all", "particle_mass"), ("all", "particle_ones")],
    ),
]


g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_ds(g30, big_data=True)
def test_particle_projection_answers():
    """

    This iterates over the all the plot modification functions in
    PROJ_ATTR_ARGS. Each time, it compares the images produced by
    ParticleProjectionPlot to the gold standard.


    """

    plot_field = ("all", "particle_mass")
    decimals = 12
    ds = data_dir_load(g30)
    for ax in "xyz":
        for attr_name in PROJ_ATTR_ARGS.keys():
            for args in PROJ_ATTR_ARGS[attr_name]:
                test = PlotWindowAttributeTest(
                    ds,
                    plot_field,
                    ax,
                    attr_name,
                    args,
                    decimals,
                    "ParticleProjectionPlot",
                )
                test_particle_projection_answers.__name__ = test.description
                yield test


@requires_ds(g30, big_data=True)
def test_particle_projection_filter():
    """

    This tests particle projection plots for filter fields.


    """

    def formed_star(pfilter, data):
        filter = data["all", "creation_time"] > 0
        return filter

    add_particle_filter(
        "formed_star",
        function=formed_star,
        filtered_type="all",
        requires=["creation_time"],
    )

    plot_field = ("formed_star", "particle_mass")

    decimals = 12
    ds = data_dir_load(g30)
    ds.add_particle_filter("formed_star")
    for ax in "xyz":
        attr_name = "set_log"
        for args in PROJ_ATTR_ARGS[attr_name]:
            test = PlotWindowAttributeTest(
                ds, plot_field, ax, attr_name, args, decimals, "ParticleProjectionPlot"
            )
            test_particle_projection_filter.__name__ = test.description
            yield test


@requires_ds(g30, big_data=True)
def test_particle_phase_answers():
    """

    This iterates over the all the plot modification functions in
    PHASE_ATTR_ARGS. Each time, it compares the images produced by
    ParticlePhasePlot to the gold standard.

    """

    decimals = 12
    ds = data_dir_load(g30)

    x_field = ("all", "particle_velocity_x")
    y_field = ("all", "particle_velocity_y")
    z_field = ("all", "particle_mass")
    for attr_name in PHASE_ATTR_ARGS.keys():
        for args in PHASE_ATTR_ARGS[attr_name]:
            test = PhasePlotAttributeTest(
                ds,
                x_field,
                y_field,
                z_field,
                attr_name,
                args,
                decimals,
                "ParticlePhasePlot",
            )

            test_particle_phase_answers.__name__ = test.description
            yield test


class TestParticlePhasePlotSave(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.curdir = os.getcwd()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)

    def test_particle_phase_plot(self):
        test_ds = fake_particle_ds()
        data_sources = [
            test_ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3),
            test_ds.all_data(),
        ]
        particle_phases = []

        for source in data_sources:
            for x_field, y_field, z_fields in PHASE_FIELDS:
                particle_phases.append(
                    ParticlePhasePlot(
                        source,
                        x_field,
                        y_field,
                        z_fields,
                        x_bins=16,
                        y_bins=16,
                    )
                )

                particle_phases.append(
                    ParticlePhasePlot(
                        source,
                        x_field,
                        y_field,
                        z_fields,
                        x_bins=16,
                        y_bins=16,
                        deposition="cic",
                    )
                )

                pp = create_profile(
                    source,
                    [x_field, y_field],
                    z_fields,
                    weight_field=("all", "particle_ones"),
                    n_bins=[16, 16],
                )

                particle_phases.append(ParticlePhasePlot.from_profile(pp))
        particle_phases[0]._repr_html_()

        with mock.patch(
            "yt.visualization._mpl_imports.FigureCanvasAgg.print_figure"
        ), mock.patch(
            "yt.visualization._mpl_imports.FigureCanvasPdf.print_figure"
        ), mock.patch(
            "yt.visualization._mpl_imports.FigureCanvasPS.print_figure"
        ):
            for p in particle_phases:
                for fname in TEST_FLNMS:
                    p.save(fname)


tgal = "TipsyGalaxy/galaxy.00300"


@requires_file(tgal)
def test_particle_phase_plot_semantics():
    ds = load(tgal)
    ad = ds.all_data()
    dens_ex = ad.quantities.extrema(("Gas", "density"))
    temp_ex = ad.quantities.extrema(("Gas", "temperature"))
    plot = ParticlePlot(
        ds, ("Gas", "density"), ("Gas", "temperature"), ("Gas", "particle_mass")
    )
    plot.set_log(("Gas", "density"), True)
    plot.set_log(("Gas", "temperature"), True)
    p = plot.profile

    # bin extrema are field extrema
    assert dens_ex[0] - np.spacing(dens_ex[0]) == p.x_bins[0]
    assert dens_ex[-1] + np.spacing(dens_ex[-1]) == p.x_bins[-1]
    assert temp_ex[0] - np.spacing(temp_ex[0]) == p.y_bins[0]
    assert temp_ex[-1] + np.spacing(temp_ex[-1]) == p.y_bins[-1]

    # bins are evenly spaced in log space
    logxbins = np.log10(p.x_bins)
    dxlogxbins = logxbins[1:] - logxbins[:-1]
    assert_allclose(dxlogxbins, dxlogxbins[0])

    logybins = np.log10(p.y_bins)
    dylogybins = logybins[1:] - logybins[:-1]
    assert_allclose(dylogybins, dylogybins[0])

    plot.set_log(("Gas", "density"), False)
    plot.set_log(("Gas", "temperature"), False)
    p = plot.profile

    # bin extrema are field extrema
    assert dens_ex[0] - np.spacing(dens_ex[0]) == p.x_bins[0]
    assert dens_ex[-1] + np.spacing(dens_ex[-1]) == p.x_bins[-1]
    assert temp_ex[0] - np.spacing(temp_ex[0]) == p.y_bins[0]
    assert temp_ex[-1] + np.spacing(temp_ex[-1]) == p.y_bins[-1]

    # bins are evenly spaced in log space
    dxbins = p.x_bins[1:] - p.x_bins[:-1]
    assert_allclose(dxbins, dxbins[0])

    dybins = p.y_bins[1:] - p.y_bins[:-1]
    assert_allclose(dybins, dybins[0])


@requires_file(tgal)
def test_set_units():
    ds = load(tgal)
    sp = ds.sphere("max", (1.0, "Mpc"))
    pp = ParticlePhasePlot(
        sp, ("Gas", "density"), ("Gas", "temperature"), ("Gas", "particle_mass")
    )
    # make sure we can set the units using the tuple without erroring out
    pp.set_unit(("Gas", "particle_mass"), "Msun")


@requires_file(tgal)
def test_switch_ds():
    """
    Tests the _switch_ds() method for ParticleProjectionPlots that as of
    25th October 2017 requires a specific hack in plot_container.py
    """
    ds = load(tgal)
    ds2 = load(tgal)

    plot = ParticlePlot(
        ds,
        ("Gas", "particle_position_x"),
        ("Gas", "particle_position_y"),
        ("Gas", "density"),
    )

    plot._switch_ds(ds2)

    return


class TestParticleProjectionPlotSave(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.curdir = os.getcwd()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        shutil.rmtree(self.tmpdir)

    def test_particle_plot(self):
        test_ds = fake_particle_ds()
        particle_projs = []
        for dim in range(3):
            particle_projs += [
                ParticleProjectionPlot(test_ds, dim, ("all", "particle_mass")),
                ParticleProjectionPlot(
                    test_ds, dim, ("all", "particle_mass"), deposition="cic"
                ),
                ParticleProjectionPlot(
                    test_ds, dim, ("all", "particle_mass"), density=True
                ),
            ]
        particle_projs[0]._repr_html_()
        with mock.patch(
            "yt.visualization._mpl_imports.FigureCanvasAgg.print_figure"
        ), mock.patch(
            "yt.visualization._mpl_imports.FigureCanvasPdf.print_figure"
        ), mock.patch(
            "yt.visualization._mpl_imports.FigureCanvasPS.print_figure"
        ):
            for p in particle_projs:
                for fname in TEST_FLNMS:
                    p.save(fname)[0]

    def test_particle_plot_ds(self):
        test_ds = fake_particle_ds()
        ds_region = test_ds.region([0.5] * 3, [0.4] * 3, [0.6] * 3)
        for dim in range(3):
            pplot_ds = ParticleProjectionPlot(
                test_ds, dim, ("all", "particle_mass"), data_source=ds_region
            )
            with mock.patch(
                "yt.visualization._mpl_imports.FigureCanvasAgg.print_figure"
            ):
                pplot_ds.save()

    def test_particle_plot_c(self):
        test_ds = fake_particle_ds()
        for center in CENTER_SPECS:
            for dim in range(3):
                pplot_c = ParticleProjectionPlot(
                    test_ds, dim, ("all", "particle_mass"), center=center
                )
                with mock.patch(
                    "yt.visualization._mpl_imports.FigureCanvasAgg.print_figure"
                ):
                    pplot_c.save()

    def test_particle_plot_wf(self):
        test_ds = fake_particle_ds()
        for dim in range(3):
            for weight_field in WEIGHT_FIELDS:
                pplot_wf = ParticleProjectionPlot(
                    test_ds, dim, ("all", "particle_mass"), weight_field=weight_field
                )
                with mock.patch(
                    "yt.visualization._mpl_imports.FigureCanvasAgg.print_figure"
                ):
                    pplot_wf.save()

    def test_creation_with_width(self):
        test_ds = fake_particle_ds()
        for width, (xlim, ylim, pwidth, _aun) in WIDTH_SPECS.items():
            plot = ParticleProjectionPlot(
                test_ds, 0, ("all", "particle_mass"), width=width
            )

            xlim = [plot.ds.quan(el[0], el[1]) for el in xlim]
            ylim = [plot.ds.quan(el[0], el[1]) for el in ylim]
            pwidth = [plot.ds.quan(el[0], el[1]) for el in pwidth]

            [assert_array_almost_equal(px, x, 14) for px, x in zip(plot.xlim, xlim)]
            [assert_array_almost_equal(py, y, 14) for py, y in zip(plot.ylim, ylim)]
            [assert_array_almost_equal(pw, w, 14) for pw, w in zip(plot.width, pwidth)]


def test_particle_plot_instance():
    """
    Tests the type of plot instance returned by ParticlePlot.

    If x_field and y_field are any combination of valid particle_position in x,
    y or z axis,then ParticleProjectionPlot instance is expected.


    """
    ds = fake_particle_ds()
    x_field = ("all", "particle_position_x")
    y_field = ("all", "particle_position_y")
    z_field = ("all", "particle_velocity_x")

    plot = ParticlePlot(ds, x_field, y_field)
    assert isinstance(plot, ParticleProjectionPlot)

    plot = ParticlePlot(ds, y_field, x_field)
    assert isinstance(plot, ParticleProjectionPlot)

    plot = ParticlePlot(ds, x_field, z_field)
    assert isinstance(plot, ParticlePhasePlot)



"""
Data structures for AdaptaHOP



"""


HEADER_ATTRIBUTES = (
    ("npart", 1, "i"),
    ("massp", 1, "f"),
    ("aexp", 1, "f"),
    ("omega_t", 1, "f"),
    ("age", 1, "f"),
    (("nhalos", "nsubs"), 2, "i"),
)

HALO_ATTRIBUTES = (
    ("npart", 1, "i"),
    ("particle_identities", -1, "i"),
    ("particle_identifier", 1, "i"),
    ("timestep", 1, "i"),
    (("level", "host_id", "first_subhalo_id", "n_subhalos", "next_subhalo_id"), 5, "i"),
    ("particle_mass", 1, "f"),
    (("raw_position_x", "raw_position_y", "raw_position_z"), 3, "f"),
    (("particle_velocity_x", "particle_velocity_y", "particle_velocity_z"), 3, "f"),
    (
        (
            "particle_angular_momentum_x",
            "particle_angular_momentum_y",
            "particle_angular_momentum_z",
        ),
        3,
        "f",
    ),
    (("r", "a", "b", "c"), 4, "f"),
    (("ek", "ep", "etot"), 3, "f"),
    ("spin", 1, "f"),
    (("virial_radius", "virial_mass", "virial_temperature", "virial_velocity"), 4, "f"),
    (("rho0", "R_c"), 2, "f"),
)



import numpy as np

from yt.testing import assert_equal, fake_amr_ds, fake_random_ds


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_mean_sum_integrate():
    for nprocs in [-1, 1, 2, 16]:
        if nprocs == -1:
            ds = fake_amr_ds(fields=("density",), units=("g/cm**3",), particles=20)
        else:
            ds = fake_random_ds(
                32, nprocs=nprocs, fields=("density",), units=("g/cm**3",), particles=20
            )
        ad = ds.all_data()

        # Sums
        q = ad.sum(("gas", "density"))

        q1 = ad.quantities.total_quantity(("gas", "density"))

        assert_equal(q, q1)

        q = ad.sum(("all", "particle_ones"))

        q1 = ad.quantities.total_quantity(("all", "particle_ones"))

        assert_equal(q, q1)

        # Weighted Averages
        w = ad.mean(("gas", "density"))

        w1 = ad.quantities.weighted_average_quantity(
            ("gas", "density"), ("index", "ones")
        )

        assert_equal(w, w1)

        w = ad.mean(("gas", "density"), weight=("gas", "density"))

        w1 = ad.quantities.weighted_average_quantity(
            ("gas", "density"), ("gas", "density")
        )

        assert_equal(w, w1)

        w = ad.mean(("all", "particle_mass"))

        w1 = ad.quantities.weighted_average_quantity(
            ("all", "particle_mass"), ("all", "particle_ones")
        )

        assert_equal(w, w1)

        w = ad.mean(("all", "particle_mass"), weight=("all", "particle_mass"))

        w1 = ad.quantities.weighted_average_quantity(
            ("all", "particle_mass"), ("all", "particle_mass")
        )

        assert_equal(w, w1)

        # Projections
        p = ad.sum(("gas", "density"), axis=0)

        p1 = ds.proj(("gas", "density"), 0, data_source=ad, method="sum")

        assert_equal(p[("gas", "density")], p1[("gas", "density")])

        # Check by axis-name
        p = ad.sum(("gas", "density"), axis="x")

        assert_equal(p[("gas", "density")], p1[("gas", "density")])

        # Now we check proper projections
        p = ad.integrate(("gas", "density"), axis=0)
        p1 = ds.proj(("gas", "density"), 0, data_source=ad)

        assert_equal(p[("gas", "density")], p1[("gas", "density")])

        # Check by axis-name
        p = ad.integrate(("gas", "density"), axis="x")

        assert_equal(p[("gas", "density")], p1[("gas", "density")])


def test_min_max():
    for nprocs in [-1, 1, 2, 16]:
        fields = ["density", "temperature"]
        units = ["g/cm**3", "K"]
        if nprocs == -1:
            ds = fake_amr_ds(fields=fields, units=units, particles=20)
        else:
            ds = fake_random_ds(
                32, nprocs=nprocs, fields=fields, units=units, particles=20
            )

        ad = ds.all_data()

        q = ad.min(("gas", "density")).v
        assert_equal(q, ad[("gas", "density")].min())

        q = ad.max(("gas", "density")).v
        assert_equal(q, ad[("gas", "density")].max())

        q = ad.min(("all", "particle_mass")).v
        assert_equal(q, ad[("all", "particle_mass")].min())

        q = ad.max(("all", "particle_mass")).v
        assert_equal(q, ad[("all", "particle_mass")].max())

        ptp = ad.ptp(("gas", "density")).v
        assert_equal(ptp, ad[("gas", "density")].max() - ad[("gas", "density")].min())

        ptp = ad.ptp(("all", "particle_mass")).v
        assert_equal(
            ptp, ad[("all", "particle_mass")].max() - ad[("all", "particle_mass")].min()
        )

        p = ad.max(("gas", "density"), axis=1)
        p1 = ds.proj(("gas", "density"), 1, data_source=ad, method="mip")
        assert_equal(p[("gas", "density")], p1[("gas", "density")])

        p = ad.max(("gas", "density"), axis="y")
        p1 = ds.proj(("gas", "density"), 1, data_source=ad, method="mip")
        assert_equal(p[("gas", "density")], p1[("gas", "density")])

        # Test that we can get multiple in a single pass

        qrho, qtemp = ad.max([("gas", "density"), ("gas", "temperature")])
        assert_equal(qrho, ad[("gas", "density")].max())
        assert_equal(qtemp, ad[("gas", "temperature")].max())

        qrho, qtemp = ad.min([("gas", "density"), ("gas", "temperature")])
        assert_equal(qrho, ad[("gas", "density")].min())
        assert_equal(qtemp, ad[("gas", "temperature")].min())


def test_argmin():
    fields = ["density", "temperature"]
    units = ["g/cm**3", "K"]
    for nprocs in [-1, 1, 2, 16]:
        if nprocs == -1:
            ds = fake_amr_ds(fields=fields, units=units)
        else:
            ds = fake_random_ds(
                32,
                nprocs=nprocs,
                fields=fields,
                units=units,
            )

        ad = ds.all_data()

        q = ad.argmin(("gas", "density"), axis=[("gas", "density")])
        assert_equal(q, ad[("gas", "density")].min())

        q1, q2 = ad.argmin(
            ("gas", "density"), axis=[("gas", "density"), ("gas", "temperature")]
        )
        mi = np.argmin(ad[("gas", "density")])
        assert_equal(q1, ad[("gas", "density")].min())
        assert_equal(q2, ad[("gas", "temperature")][mi])

        pos = ad.argmin(("gas", "density"))
        mi = np.argmin(ad[("gas", "density")])
        assert_equal(pos[0], ad[("index", "x")][mi])
        assert_equal(pos[1], ad[("index", "y")][mi])
        assert_equal(pos[2], ad[("index", "z")][mi])


def test_argmax():
    fields = ["density", "temperature"]
    units = ["g/cm**3", "K"]
    for nprocs in [-1, 1, 2, 16]:
        if nprocs == -1:
            ds = fake_amr_ds(fields=fields, units=units)
        else:
            ds = fake_random_ds(
                32,
                nprocs=nprocs,
                fields=fields,
                units=units,
            )

        ad = ds.all_data()

        q = ad.argmax(("gas", "density"), axis=[("gas", "density")])
        assert_equal(q, ad[("gas", "density")].max())

        q1, q2 = ad.argmax(
            ("gas", "density"), axis=[("gas", "density"), ("gas", "temperature")]
        )
        mi = np.argmax(ad[("gas", "density")])
        assert_equal(q1, ad[("gas", "density")].max())
        assert_equal(q2, ad[("gas", "temperature")][mi])

        pos = ad.argmax(("gas", "density"))
        mi = np.argmax(ad[("gas", "density")])
        assert_equal(pos[0], ad[("index", "x")][mi])
        assert_equal(pos[1], ad[("index", "y")][mi])
        assert_equal(pos[2], ad[("index", "z")][mi])



# from . import species_fields
from . import (
    angular_momentum,
    astro_fields,
    cosmology_fields,
    fluid_fields,
    fluid_vector_fields,
    geometric_fields,
    local_fields,
    magnetic_field,
    my_plugin_fields,
    particle_fields,
    vector_operations,
)
from .derived_field import (
    DerivedField,
    ValidateDataField,
    ValidateGridType,
    ValidateParameter,
    ValidateProperty,
    ValidateSpatial,
)
from .field_detector import FieldDetector
from .field_info_container import FieldInfoContainer
from .field_plugin_registry import field_plugins, register_field_plugin
from .local_fields import add_field, derived_field
from .xray_emission_fields import add_xray_emissivity_field



import pytest

from yt.visualization._commons import validate_image_name


@pytest.mark.parametrize(
    "name, expected",
    [
        ("noext", "noext.png"),
        ("nothing.png", "nothing.png"),
        ("nothing.pdf", "nothing.pdf"),
        ("version.1.2.3", "version.1.2.3.png"),
    ],
)
def test_default(name, expected):
    result = validate_image_name(name)
    assert result == expected


@pytest.mark.parametrize(
    "name, suffix, expected",
    [
        ("noext", ".png", "noext.png"),
        ("nothing.png", ".png", "nothing.png"),
        ("nothing.png", ".pdf", "nothing.pdf"),
        ("nothing.pdf", ".pdf", "nothing.pdf"),
        ("nothing.pdf", ".png", "nothing.png"),
        ("version.1.2.3", ".png", "version.1.2.3.png"),
        ("version.1.2.3", ".pdf", "version.1.2.3.pdf"),
    ],
)
def test_custom_valid_ext(name, suffix, expected):
    alt_suffix = suffix.replace(".", "")
    result1 = validate_image_name(name, suffix=suffix)
    result2 = validate_image_name(name, suffix=alt_suffix)

    assert result1 == expected
    assert result2 == expected



import os
from collections import defaultdict

import numpy as np

from yt.frontends.chombo.io import parse_orion_sinks
from yt.funcs import mylog
from yt.geometry.selection_routines import GridSelector
from yt.utilities.io_handler import BaseIOHandler


def _remove_raw(all_fields, raw_fields):
    centered_fields = set(all_fields)
    for raw in raw_fields:
        centered_fields.discard(raw)
    return list(centered_fields)


class IOHandlerBoxlib(BaseIOHandler):

    _dataset_type = "boxlib_native"

    def __init__(self, ds, *args, **kwargs):
        super().__init__(ds)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        if any((not (ftype == "boxlib" or ftype == "raw") for ftype, fname in fields)):
            raise NotImplementedError
        rv = {}
        raw_fields = []
        for field in fields:
            if field[0] == "raw":
                nodal_flag = self.ds.nodal_flags[field[1]]
                num_nodes = 2 ** sum(nodal_flag)
                rv[field] = np.empty((size, num_nodes), dtype="float64")
                raw_fields.append(field)
            else:
                rv[field] = np.empty(size, dtype="float64")
        centered_fields = _remove_raw(fields, raw_fields)
        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s grids",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )
        ind = 0
        for chunk in chunks:
            data = self._read_chunk_data(chunk, centered_fields)
            for g in chunk.objs:
                for field in fields:
                    if field in centered_fields:
                        ds = data[g.id].pop(field)
                    else:
                        ds = self._read_raw_field(g, field)
                    nd = g.select(selector, ds, rv[field], ind)
                ind += nd
                data.pop(g.id)
        return rv

    def _read_raw_field(self, grid, field):
        field_name = field[1]
        base_dir = self.ds.index.raw_file

        nghost = self.ds.index.raw_field_nghost[field_name]
        box_list = self.ds.index.raw_field_map[field_name][0]
        fn_list = self.ds.index.raw_field_map[field_name][1]
        offset_list = self.ds.index.raw_field_map[field_name][2]

        lev = grid.Level
        filename = base_dir + "Level_%d/" % lev + fn_list[grid.id]
        offset = offset_list[grid.id]
        box = box_list[grid.id]

        lo = box[0] - nghost
        hi = box[1] + nghost
        shape = hi - lo + 1
        with open(filename, "rb") as f:
            f.seek(offset)
            f.readline()  # always skip the first line
            arr = np.fromfile(f, "float64", np.product(shape))
            arr = arr.reshape(shape, order="F")
        return arr[
            tuple(
                slice(None) if (nghost[dim] == 0) else slice(nghost[dim], -nghost[dim])
                for dim in range(self.ds.dimensionality)
            )
        ]

    def _read_chunk_data(self, chunk, fields):
        data = {}
        grids_by_file = defaultdict(list)
        if len(chunk.objs) == 0:
            return data
        for g in chunk.objs:
            if g.filename is None:
                continue
            grids_by_file[g.filename].append(g)
        dtype = self.ds.index._dtype
        bpr = dtype.itemsize
        for filename in grids_by_file:
            grids = grids_by_file[filename]
            grids.sort(key=lambda a: a._offset)
            f = open(filename, "rb")
            for grid in grids:
                data[grid.id] = {}
                local_offset = grid._get_offset(f) - f.tell()
                count = grid.ActiveDimensions.prod()
                size = count * bpr
                for field in self.ds.index.field_order:
                    if field in fields:
                        # We read it ...
                        f.seek(local_offset, os.SEEK_CUR)
                        v = np.fromfile(f, dtype=dtype, count=count)
                        v = v.reshape(grid.ActiveDimensions, order="F")
                        data[grid.id][field] = v
                        local_offset = 0
                    else:
                        local_offset += size
        return data

    def _read_particle_coords(self, chunks, ptf):
        yield from self._read_particle_fields(chunks, ptf, None)

    def _read_particle_fields(self, chunks, ptf, selector):
        for chunk in chunks:  # These should be organized by grid filename
            for g in chunk.objs:
                for ptype, field_list in sorted(ptf.items()):
                    npart = g._pdata[ptype]["NumberOfParticles"]
                    if npart == 0:
                        continue

                    fn = g._pdata[ptype]["particle_filename"]
                    offset = g._pdata[ptype]["offset"]
                    pheader = self.ds.index.particle_headers[ptype]

                    with open(fn, "rb") as f:
                        # read in the position fields for selection
                        f.seek(offset + pheader.particle_int_dtype.itemsize * npart)
                        rdata = np.fromfile(
                            f, pheader.real_type, pheader.num_real * npart
                        )
                        x = np.asarray(rdata[0 :: pheader.num_real], dtype=np.float64)
                        y = np.asarray(rdata[1 :: pheader.num_real], dtype=np.float64)
                        if g.ds.dimensionality == 2:
                            z = np.ones_like(y)
                            z *= 0.5 * (g.LeftEdge[2] + g.RightEdge[2])
                        else:
                            z = np.asarray(
                                rdata[2 :: pheader.num_real], dtype=np.float64
                            )

                        if selector is None:
                            # This only ever happens if the call is made from
                            # _read_particle_coords.
                            yield ptype, (x, y, z)
                            continue
                        mask = selector.select_points(x, y, z, 0.0)
                        if mask is None:
                            continue
                        for field in field_list:
                            # handle the case that this is an integer field
                            int_fnames = [
                                fname for _, fname in pheader.known_int_fields
                            ]
                            if field in int_fnames:
                                ind = int_fnames.index(field)
                                f.seek(offset)
                                idata = np.fromfile(
                                    f, pheader.int_type, pheader.num_int * npart
                                )
                                data = np.asarray(
                                    idata[ind :: pheader.num_int], dtype=np.float64
                                )
                                yield (ptype, field), data[mask].flatten()

                            # handle case that this is a real field
                            real_fnames = [
                                fname for _, fname in pheader.known_real_fields
                            ]
                            if field in real_fnames:
                                ind = real_fnames.index(field)
                                data = np.asarray(
                                    rdata[ind :: pheader.num_real], dtype=np.float64
                                )
                                yield (ptype, field), data[mask].flatten()


class IOHandlerOrion(IOHandlerBoxlib):
    _dataset_type = "orion_native"

    _particle_filename = None

    @property
    def particle_filename(self):
        fn = self.ds.output_dir + "/StarParticles"
        if not os.path.exists(fn):
            fn = self.ds.output_dir + "/SinkParticles"
        self._particle_filename = fn
        return self._particle_filename

    _particle_field_index = None

    @property
    def particle_field_index(self):

        index = parse_orion_sinks(self.particle_filename)

        self._particle_field_index = index
        return self._particle_field_index

    def _read_particle_selection(self, chunks, selector, fields):
        rv = {}
        chunks = list(chunks)

        if isinstance(selector, GridSelector):

            if not (len(chunks) == len(chunks[0].objs) == 1):
                raise RuntimeError

            grid = chunks[0].objs[0]

            for ftype, fname in fields:
                rv[ftype, fname] = self._read_particles(grid, fname)

            return rv

        rv = {f: np.array([]) for f in fields}
        for chunk in chunks:
            for grid in chunk.objs:
                for ftype, fname in fields:
                    data = self._read_particles(grid, fname)
                    rv[ftype, fname] = np.concatenate((data, rv[ftype, fname]))
        return rv

    def _read_particles(self, grid, field):
        """
        parses the Orion Star Particle text files

        """

        particles = []

        if grid.NumberOfParticles == 0:
            return np.array(particles)

        def read(line, field):
            entry = line.strip().split(" ")[self.particle_field_index[field]]
            return float(entry)

        try:
            lines = self._cached_lines
            for num in grid._particle_line_numbers:
                line = lines[num]
                particles.append(read(line, field))
            return np.array(particles)
        except AttributeError:
            fn = self.particle_filename
            with open(fn) as f:
                lines = f.readlines()
                self._cached_lines = lines
                for num in grid._particle_line_numbers:
                    line = lines[num]
                    particles.append(read(line, field))
            return np.array(particles)



import numpy as np

from yt import load
from yt.frontends.stream.fields import StreamFieldInfo
from yt.testing import (
    assert_allclose_units,
    assert_almost_equal,
    assert_array_almost_equal_nulp,
    assert_array_equal,
    assert_equal,
    assert_raises,
    fake_amr_ds,
    fake_particle_ds,
    fake_random_ds,
    requires_file,
)
from yt.units.yt_array import YTArray, YTQuantity, array_like_field
from yt.utilities.cosmology import Cosmology
from yt.utilities.exceptions import (
    YTDimensionalityError,
    YTFieldUnitError,
    YTFieldUnitParseError,
)


def get_params(ds):
    return dict(
        axis=0,
        center=YTArray((0.0, 0.0, 0.0), "cm", registry=ds.unit_registry),
        bulk_velocity=YTArray((0.0, 0.0, 0.0), "cm/s", registry=ds.unit_registry),
        bulk_magnetic_field=YTArray((0.0, 0.0, 0.0), "G", registry=ds.unit_registry),
        normal=YTArray((0.0, 0.0, 1.0), "", registry=ds.unit_registry),
        cp_x_vec=YTArray((1.0, 0.0, 0.0), "", registry=ds.unit_registry),
        cp_y_vec=YTArray((0.0, 1.0, 0.0), "", registry=ds.unit_registry),
        cp_z_vec=YTArray((0.0, 0.0, 1.0), "", registry=ds.unit_registry),
        omega_baryon=0.04,
        observer_redshift=0.0,
        source_redshift=3.0,
        virial_radius=YTQuantity(1.0, "cm"),
    )


_base_fields = (
    ("gas", "density"),
    ("gas", "velocity_x"),
    ("gas", "velocity_y"),
    ("gas", "velocity_z"),
)


def _strip_ftype(field):
    if not isinstance(field, tuple):
        return field
    elif field[0] in ("all", "io"):
        return field
    return field[1]


class TestFieldAccess:
    description = None

    def __init__(self, field_name, ds, nprocs):
        # Note this should be a field name
        self.field_name = field_name
        self.description = f"Accessing_{field_name}_{nprocs}"
        self.nprocs = nprocs
        self.ds = ds

    def __call__(self):
        field = self.ds._get_field_info(*self.field_name)
        skip_grids = False
        needs_spatial = False
        for v in field.validators:
            if getattr(v, "ghost_zones", 0) > 0:
                skip_grids = True
            if hasattr(v, "ghost_zones"):
                needs_spatial = True

        ds = self.ds

        # This gives unequal sized grids as well as subgrids
        dd1 = ds.all_data()
        dd2 = ds.all_data()
        sp = get_params(ds)
        dd1.field_parameters.update(sp)
        dd2.field_parameters.update(sp)
        with np.errstate(all="ignore"):
            v1 = dd1[self.field_name]
            # No more conversion checking
            assert_equal(v1, dd1[self.field_name])
            if not needs_spatial:
                with field.unit_registry(dd2):
                    res = field._function(field, dd2)
                    res = dd2.apply_units(res, field.units)
                assert_array_almost_equal_nulp(v1, res, 4)
            if not skip_grids:
                for g in ds.index.grids:
                    g.field_parameters.update(sp)
                    v1 = g[self.field_name]
                    g.clear_data()
                    g.field_parameters.update(sp)
                    r1 = field._function(field, g)
                    if field.sampling_type == "particle":
                        assert_equal(v1.shape[0], g.NumberOfParticles)
                    else:
                        assert_array_equal(r1.shape, v1.shape)
                        for ax in "xyz":
                            assert_array_equal(g["index", ax].shape, v1.shape)
                    with field.unit_registry(g):
                        res = field._function(field, g)
                        assert_array_equal(v1.shape, res.shape)
                        res = g.apply_units(res, field.units)
                    assert_array_almost_equal_nulp(v1, res, 4)


def get_base_ds(nprocs):
    fields, units = [], []

    for fname, (code_units, *_) in StreamFieldInfo.known_other_fields:
        fields.append(("gas", fname))
        units.append(code_units)

    pfields, punits = [], []

    for fname, (code_units, _aliases, _dn) in StreamFieldInfo.known_particle_fields:
        if fname == "smoothing_lenth":
            # we test SPH fields elsewhere
            continue
        pfields.append(fname)
        punits.append(code_units)

    ds = fake_random_ds(
        4,
        fields=fields,
        units=units,
        particles=20,
        nprocs=nprocs,
        particle_fields=pfields,
        particle_field_units=punits,
    )
    ds.parameters["HydroMethod"] = "streaming"
    ds.parameters["EOSType"] = 1.0
    ds.parameters["EOSSoundSpeed"] = 1.0
    ds.conversion_factors["Time"] = 1.0
    ds.conversion_factors.update({f: 1.0 for f in fields})
    ds.gamma = 5.0 / 3.0
    ds.current_redshift = 0.0001
    ds.cosmological_simulation = 1
    ds.hubble_constant = 0.7
    ds.omega_matter = 0.27
    ds.omega_lambda = 0.73
    ds.cosmology = Cosmology(
        hubble_constant=ds.hubble_constant,
        omega_matter=ds.omega_matter,
        omega_lambda=ds.omega_lambda,
        unit_registry=ds.unit_registry,
    )
    # ensures field errors are raised during testing
    # see FieldInfoContainer.check_derived_fields
    ds._field_test_dataset = True
    ds.index
    return ds


def test_all_fields():
    datasets = {}

    for nprocs in [1, 4, 8]:
        ds = get_base_ds(nprocs)
        datasets[nprocs] = ds

    for field in sorted(ds.field_info):
        if field[1].find("beta_p") > -1:
            continue
        if field[1].find("vertex") > -1:
            # don't test the vertex fields for now
            continue
        if field[1].find("smoothed") > -1:
            # smoothed fields aren't implemented for grid data
            continue
        if field in ds.field_list:
            # Don't know how to test this.  We need some way of having fields
            # that are fallbacks be tested, but we don't have that now.
            continue

        for nprocs in [1, 4, 8]:
            test_all_fields.__name__ = f"{field}_{nprocs}"
            yield TestFieldAccess(field, datasets[nprocs], nprocs)


def test_add_deposited_particle_field():
    # NOT tested: "std", "mesh_id", "nearest" and "simple_smooth"
    base_ds = get_base_ds(1)
    ad = base_ds.all_data()

    # Test "count", "sum" and "cic" method
    for method in ["count", "sum", "cic"]:
        fn = base_ds.add_deposited_particle_field(("io", "particle_mass"), method)
        expected_fn = "io_%s" if method == "count" else "io_%s_mass"
        assert_equal(fn, ("deposit", expected_fn % method))
        ret = ad[fn]
        if method == "count":
            assert_equal(ret.sum(), ad[("io", "particle_ones")].sum())
        else:
            assert_almost_equal(ret.sum(), ad[("io", "particle_mass")].sum())

    # Test "weighted_mean" method
    fn = base_ds.add_deposited_particle_field(
        ("io", "particle_ones"), "weighted_mean", weight_field="particle_ones"
    )
    assert_equal(fn, ("deposit", "io_avg_ones"))
    ret = ad[fn]
    # The sum should equal the number of cells that have particles
    assert_equal(ret.sum(), np.count_nonzero(ad[("deposit", "io_count")]))


def test_add_gradient_fields():
    ds = get_base_ds(1)
    gfields = ds.add_gradient_fields(("gas", "density"))
    gfields += ds.add_gradient_fields(("index", "ones"))
    field_list = [
        ("gas", "density_gradient_x"),
        ("gas", "density_gradient_y"),
        ("gas", "density_gradient_z"),
        ("gas", "density_gradient_magnitude"),
        ("index", "ones_gradient_x"),
        ("index", "ones_gradient_y"),
        ("index", "ones_gradient_z"),
        ("index", "ones_gradient_magnitude"),
    ]
    assert_equal(gfields, field_list)
    ad = ds.all_data()
    for field in field_list:
        ret = ad[field]
        if field[0] == "gas":
            assert str(ret.units) == "g/cm**4"
        else:
            assert str(ret.units) == "1/cm"


def test_add_gradient_fields_by_fname():
    ds = fake_amr_ds(fields=("density", "temperature"), units=("g/cm**3", "K"))
    actual = ds.add_gradient_fields(("gas", "density"))
    expected = [
        ("gas", "density_gradient_x"),
        ("gas", "density_gradient_y"),
        ("gas", "density_gradient_z"),
        ("gas", "density_gradient_magnitude"),
    ]
    assert_equal(actual, expected)


def test_add_gradient_multiple_fields():
    ds = fake_amr_ds(fields=("density", "temperature"), units=("g/cm**3", "K"))
    actual = ds.add_gradient_fields([("gas", "density"), ("gas", "temperature")])
    expected = [
        ("gas", "density_gradient_x"),
        ("gas", "density_gradient_y"),
        ("gas", "density_gradient_z"),
        ("gas", "density_gradient_magnitude"),
        ("gas", "temperature_gradient_x"),
        ("gas", "temperature_gradient_y"),
        ("gas", "temperature_gradient_z"),
        ("gas", "temperature_gradient_magnitude"),
    ]
    assert_equal(actual, expected)

    ds = fake_amr_ds(fields=("density", "temperature"), units=("g/cm**3", "K"))
    actual = ds.add_gradient_fields([("gas", "density"), ("gas", "temperature")])
    assert_equal(actual, expected)


def test_add_gradient_fields_curvilinear():
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"], geometry="spherical")
    gfields = ds.add_gradient_fields(("gas", "density"))
    gfields += ds.add_gradient_fields(("index", "ones"))
    field_list = [
        ("gas", "density_gradient_r"),
        ("gas", "density_gradient_theta"),
        ("gas", "density_gradient_phi"),
        ("gas", "density_gradient_magnitude"),
        ("index", "ones_gradient_r"),
        ("index", "ones_gradient_theta"),
        ("index", "ones_gradient_phi"),
        ("index", "ones_gradient_magnitude"),
    ]
    assert_equal(gfields, field_list)
    ad = ds.all_data()
    for field in field_list:
        ret = ad[field]
        if field[0] == "gas":
            assert str(ret.units) == "g/cm**4"
        else:
            assert str(ret.units) == "1/cm"


def get_data(ds, field_name):
    # Need to create a new data object otherwise the errors we are
    # intentionally raising lead to spurious GenerationInProgress errors
    ad = ds.all_data()
    return ad[field_name]


def test_add_field_unit_semantics():
    ds = fake_random_ds(16)
    ad = ds.all_data()

    def density_alias(field, data):
        return data[("gas", "density")].in_cgs()

    def unitless_data(field, data):
        return np.ones(data[("gas", "density")].shape)

    ds.add_field(
        ("gas", "density_alias_no_units"), sampling_type="cell", function=density_alias
    )
    ds.add_field(
        ("gas", "density_alias_auto"),
        sampling_type="cell",
        function=density_alias,
        units="auto",
        dimensions="density",
    )
    ds.add_field(
        ("gas", "density_alias_wrong_units"),
        function=density_alias,
        sampling_type="cell",
        units="m/s",
    )
    ds.add_field(
        ("gas", "density_alias_unparseable_units"),
        sampling_type="cell",
        function=density_alias,
        units="dragons",
    )
    ds.add_field(
        ("gas", "density_alias_auto_wrong_dims"),
        function=density_alias,
        sampling_type="cell",
        units="auto",
        dimensions="temperature",
    )
    assert_raises(YTFieldUnitError, get_data, ds, ("gas", "density_alias_no_units"))
    assert_raises(YTFieldUnitError, get_data, ds, ("gas", "density_alias_wrong_units"))
    assert_raises(
        YTFieldUnitParseError, get_data, ds, ("gas", "density_alias_unparseable_units")
    )
    assert_raises(
        YTDimensionalityError, get_data, ds, ("gas", "density_alias_auto_wrong_dims")
    )

    dens = ad[("gas", "density_alias_auto")]
    assert_equal(str(dens.units), "g/cm**3")

    ds.add_field(("gas", "dimensionless"), sampling_type="cell", function=unitless_data)
    ds.add_field(
        ("gas", "dimensionless_auto"),
        function=unitless_data,
        sampling_type="cell",
        units="auto",
        dimensions="dimensionless",
    )
    ds.add_field(
        ("gas", "dimensionless_explicit"),
        function=unitless_data,
        sampling_type="cell",
        units="",
    )
    ds.add_field(
        ("gas", "dimensionful"),
        sampling_type="cell",
        function=unitless_data,
        units="g/cm**3",
    )

    assert_equal(str(ad[("gas", "dimensionless")].units), "dimensionless")
    assert_equal(str(ad[("gas", "dimensionless_auto")].units), "dimensionless")
    assert_equal(str(ad[("gas", "dimensionless_explicit")].units), "dimensionless")
    assert_raises(YTFieldUnitError, get_data, ds, ("gas", "dimensionful"))


def test_array_like_field():
    ds = fake_random_ds(4, particles=64)
    ad = ds.all_data()
    u1 = ad[("all", "particle_mass")].units
    u2 = array_like_field(ad, 1.0, ("all", "particle_mass")).units
    assert u1 == u2


ISOGAL = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_file(ISOGAL)
def test_array_like_field_output_units():
    ds = load(ISOGAL)
    ad = ds.all_data()
    u1 = ad[("all", "particle_mass")].units
    u2 = array_like_field(ad, 1.0, ("all", "particle_mass")).units
    assert u1 == u2
    assert str(u1) == ds.fields.all.particle_mass.output_units
    u1 = ad[("gas", "x")].units
    u2 = array_like_field(ad, 1.0, ("gas", "x")).units
    assert u1 == u2
    assert str(u1) == ds.fields.gas.x.units


def test_add_field_string():
    ds = fake_random_ds(16)
    ad = ds.all_data()

    def density_alias(field, data):
        return data[("gas", "density")]

    ds.add_field(
        ("gas", "density_alias"),
        sampling_type="cell",
        function=density_alias,
        units="g/cm**3",
    )

    ad[("gas", "density_alias")]

    assert ("gas", "density_alias") in ds.derived_field_list


def test_add_field_string_aliasing():
    ds = fake_random_ds(16)

    def density_alias(field, data):
        return data["gas", "density"]

    ds.add_field(
        ("gas", "density_alias"),
        sampling_type="cell",
        function=density_alias,
        units="g/cm**3",
    )

    ds.field_info["gas", "density_alias"]
    ds.field_info["gas", "density_alias"]

    ds = fake_particle_ds()

    def pmass_alias(field, data):
        return data["all", "particle_mass"]

    ds.add_field(
        ("all", "particle_mass_alias"),
        function=pmass_alias,
        units="g",
        sampling_type="particle",
    )

    ds.field_info["all", "particle_mass_alias"]
    ds.field_info["all", "particle_mass_alias"]


def test_morton_index():
    ds = fake_amr_ds()
    mi = ds.r["index", "morton_index"]
    mi2 = mi.view("uint64")
    assert_equal(np.unique(mi2).size, mi2.size)
    a1 = np.argsort(mi)
    a2 = np.argsort(mi2)
    assert_array_equal(a1, a2)


def test_field_inference():
    ds = fake_random_ds(16)
    ds.index
    # If this is not true this means the result of field inference depends
    # on the order we did field detection, which is random in Python3
    assert_equal(ds._last_freq, (None, None))


@requires_file(ISOGAL)
def test_deposit_amr():
    ds = load(ISOGAL)
    for g in ds.index.grids:
        gpm = g["all", "particle_mass"].sum()
        dpm = g["deposit", "all_mass"].sum()
        assert_allclose_units(gpm, dpm)


def test_ion_field_labels():
    fields = [
        "O_p1_number_density",
        "O2_p1_number_density",
        "CO2_p1_number_density",
        "Co_p1_number_density",
        "O2_p2_number_density",
        "H2O_p1_number_density",
    ]
    units = ["cm**-3" for f in fields]
    ds = fake_random_ds(16, fields=fields, units=units)

    # by default labels should use roman numerals
    default_labels = {
        "O_p1_number_density": "$\\rm{O\\ II\\ Number\\ Density}$",
        "O2_p1_number_density": "$\\rm{O_{2}\\ II\\ Number\\ Density}$",
        "CO2_p1_number_density": "$\\rm{CO_{2}\\ II\\ Number\\ Density}$",
        "Co_p1_number_density": "$\\rm{Co\\ II\\ Number\\ Density}$",
        "O2_p2_number_density": "$\\rm{O_{2}\\ III\\ Number\\ Density}$",
        "H2O_p1_number_density": "$\\rm{H_{2}O\\ II\\ Number\\ Density}$",
    }

    pm_labels = {
        "O_p1_number_density": "$\\rm{{O}^{+}\\ Number\\ Density}$",
        "O2_p1_number_density": "$\\rm{{O_{2}}^{+}\\ Number\\ Density}$",
        "CO2_p1_number_density": "$\\rm{{CO_{2}}^{+}\\ Number\\ Density}$",
        "Co_p1_number_density": "$\\rm{{Co}^{+}\\ Number\\ Density}$",
        "O2_p2_number_density": "$\\rm{{O_{2}}^{++}\\ Number\\ Density}$",
        "H2O_p1_number_density": "$\\rm{{H_{2}O}^{+}\\ Number\\ Density}$",
    }

    fobj = ds.fields.stream

    for f in fields:
        label = getattr(fobj, f).get_latex_display_name()
        assert_equal(label, default_labels[f])

    ds.set_field_label_format("ionization_label", "plus_minus")
    fobj = ds.fields.stream

    for f in fields:
        label = getattr(fobj, f).get_latex_display_name()
        assert_equal(label, pm_labels[f])



from collections import defaultdict

import numpy as np

from yt.units.yt_array import YTArray
from yt.utilities.io_handler import io_registry

from .field_exceptions import NeedsGridType

fp_units = {
    "bulk_velocity": "cm/s",
    "center": "cm",
    "normal": "",
    "cp_x_vec": "",
    "cp_y_vec": "",
    "cp_z_vec": "",
    "x_hat": "",
    "y_hat": "",
    "z_hat": "",
    "omega_baryon": "",
    "virial_radius": "cm",
    "observer_redshift": "",
    "source_redshift": "",
}


class FieldDetector(defaultdict):
    Level = 1
    NumberOfParticles = 1
    _read_exception = None
    _id_offset = 0
    domain_id = 0

    def __init__(self, nd=16, ds=None, flat=False, field_parameters=None):
        self.nd = nd
        self.flat = flat
        self._spatial = not flat
        self.ActiveDimensions = [nd, nd, nd]
        self.shape = tuple(self.ActiveDimensions)
        self.size = np.prod(self.ActiveDimensions)
        self.LeftEdge = [0.0, 0.0, 0.0]
        self.RightEdge = [1.0, 1.0, 1.0]
        self.dds = np.ones(3, "float64")
        if field_parameters is None:
            self.field_parameters = {}
        else:
            self.field_parameters = field_parameters

        class fake_dataset(defaultdict):
            pass

        if ds is None:
            # required attrs
            ds = fake_dataset(lambda: 1)
            ds["Massarr"] = np.ones(6)
            ds.current_redshift = 0.0
            ds.omega_lambda = 0.0
            ds.omega_matter = 0.0
            ds.cosmological_simulation = 0
            ds.gamma = 5.0 / 3.0
            ds.hubble_constant = 0.7
            ds.domain_left_edge = np.zeros(3, "float64")
            ds.domain_right_edge = np.ones(3, "float64")
            ds.dimensionality = 3
            ds.force_periodicity()
        self.ds = ds

        class fake_index:
            class fake_io:
                def _read_data_set(io_self, data, field):
                    return self._read_data(field)

                _read_exception = RuntimeError

            io = fake_io()

            def get_smallest_dx(self):
                return 1.0

        self.index = fake_index()
        self.requested = []
        self.requested_parameters = []
        if not self.flat:
            defaultdict.__init__(
                self,
                lambda: np.ones((nd, nd, nd), dtype="float64")
                + 1e-4 * np.random.random((nd, nd, nd)),
            )
        else:
            defaultdict.__init__(
                self,
                lambda: np.ones((nd * nd * nd), dtype="float64")
                + 1e-4 * np.random.random(nd * nd * nd),
            )

    def _reshape_vals(self, arr):
        if not self._spatial:
            return arr
        if len(arr.shape) == 3:
            return arr
        return arr.reshape(self.ActiveDimensions, order="C")

    def __missing__(self, item):
        if not isinstance(item, tuple):
            field = ("unknown", item)
        else:
            field = item
        finfo = self.ds._get_field_info(*field)
        params, permute_params = finfo._get_needed_parameters(self)
        self.field_parameters.update(params)
        # For those cases where we are guessing the field type, we will
        # need to re-update -- otherwise, our item will always not have the
        # field type.  This can lead to, for instance, "unknown" particle
        # types not getting correctly identified.
        # Note that the *only* way this works is if we also fix our field
        # dependencies during checking.  Bug #627 talks about this.
        item = self.ds._last_freq
        if finfo is not None and finfo._function.__name__ != "NullFunc":
            try:
                for param, param_v in permute_params.items():
                    for v in param_v:
                        self.field_parameters[param] = v
                        vv = finfo(self)
                if not permute_params:
                    vv = finfo(self)
            except NeedsGridType as exc:
                ngz = exc.ghost_zones
                nfd = FieldDetector(
                    self.nd + ngz * 2,
                    ds=self.ds,
                    field_parameters=self.field_parameters.copy(),
                )
                nfd._num_ghost_zones = ngz
                vv = finfo(nfd)
                if ngz > 0:
                    vv = vv[ngz:-ngz, ngz:-ngz, ngz:-ngz]
                for i in nfd.requested:
                    if i not in self.requested:
                        self.requested.append(i)
                for i in nfd.requested_parameters:
                    if i not in self.requested_parameters:
                        self.requested_parameters.append(i)
            if vv is not None:
                if not self.flat:
                    self[item] = vv
                else:
                    self[item] = vv.ravel()
                return self[item]
        elif finfo is not None and finfo.sampling_type == "particle":
            io = io_registry[self.ds.dataset_type](self.ds)
            if hasattr(io, "_vector_fields") and (
                item in io._vector_fields or item[1] in io._vector_fields
            ):
                try:
                    cols = io._vector_fields[item]
                except KeyError:
                    cols = io._vector_fields[item[1]]
                # A vector
                self[item] = YTArray(
                    np.ones((self.NumberOfParticles, cols)),
                    finfo.units,
                    registry=self.ds.unit_registry,
                )
            else:
                # Not a vector
                self[item] = YTArray(
                    np.ones(self.NumberOfParticles),
                    finfo.units,
                    registry=self.ds.unit_registry,
                )
            if item == ("STAR", "BIRTH_TIME"):
                # hack for the artio frontend so we pass valid times to
                # the artio functions for calculating physical times
                # from internal times
                self[item] *= -0.1
            self.requested.append(item)
            return self[item]
        self.requested.append(item)
        if item not in self:
            self[item] = self._read_data(item)
        return self[item]

    def _debug(self):
        # We allow this to pass through.
        return

    def deposit(self, *args, **kwargs):
        from yt.data_objects.static_output import ParticleDataset
        from yt.frontends.stream.data_structures import StreamParticlesDataset

        if kwargs["method"] == "mesh_id":
            if isinstance(self.ds, (StreamParticlesDataset, ParticleDataset)):
                raise ValueError
        return np.random.random((self.nd, self.nd, self.nd))

    def mesh_sampling_particle_field(self, *args, **kwargs):
        pos = args[0]
        npart = len(pos)
        return np.random.rand(npart)

    def smooth(self, *args, **kwargs):
        tr = np.random.random((self.nd, self.nd, self.nd))
        if kwargs["method"] == "volume_weighted":
            return [tr]

        return tr

    def particle_operation(self, *args, **kwargs):
        return None

    def _read_data(self, field_name):
        self.requested.append(field_name)
        finfo = self.ds._get_field_info(*field_name)
        if finfo.sampling_type == "particle":
            self.requested.append(field_name)
            return np.ones(self.NumberOfParticles)
        return YTArray(
            defaultdict.__missing__(self, field_name),
            units=finfo.units,
            registry=self.ds.unit_registry,
        )

    def get_field_parameter(self, param, default=0.0):
        if self.field_parameters and param in self.field_parameters:
            return self.field_parameters[param]
        self.requested_parameters.append(param)
        if param in ["center", "normal"] or param.startswith("bulk"):
            if param == "bulk_magnetic_field":
                if self.ds.unit_system.has_current_mks:
                    unit = "T"
                else:
                    unit = "G"
            else:
                unit = fp_units[param]
            return self.ds.arr(np.random.random(3) * 1e-2, unit)
        elif param in ["surface_height"]:
            return self.ds.quan(0.0, "code_length")
        elif param in ["axis"]:
            return 0
        elif param.startswith("cp_"):
            ax = param[3]
            rv = self.ds.arr((0.0, 0.0, 0.0), fp_units[param])
            rv["xyz".index(ax)] = 1.0
            return rv
        elif param.endswith("_hat"):
            ax = param[0]
            rv = YTArray((0.0, 0.0, 0.0), fp_units[param])
            rv["xyz".index(ax)] = 1.0
            return rv
        elif param == "fof_groups":
            return None
        elif param == "mu":
            return 1.0
        else:
            return default

    _num_ghost_zones = 0
    id = 1

    def apply_units(self, arr, units):
        return self.ds.arr(arr, units=units)

    def has_field_parameter(self, param):
        return param in self.field_parameters

    @property
    def fcoords(self):
        fc = np.array(
            np.mgrid[0 : 1 : self.nd * 1j, 0 : 1 : self.nd * 1j, 0 : 1 : self.nd * 1j]
        )
        if self.flat:
            fc.shape = (self.nd * self.nd * self.nd, 3)
        else:
            fc = fc.transpose()
        return self.ds.arr(fc, units="code_length")

    @property
    def fcoords_vertex(self):
        fc = np.random.random((self.nd, self.nd, self.nd, 8, 3))
        if self.flat:
            fc.shape = (self.nd * self.nd * self.nd, 8, 3)
        return self.ds.arr(fc, units="code_length")

    @property
    def icoords(self):
        ic = np.mgrid[
            0 : self.nd - 1 : self.nd * 1j,
            0 : self.nd - 1 : self.nd * 1j,
            0 : self.nd - 1 : self.nd * 1j,
        ]
        if self.flat:
            ic.shape = (self.nd * self.nd * self.nd, 3)
        else:
            ic = ic.transpose()
        return ic

    @property
    def ires(self):
        ir = np.ones(self.nd ** 3, dtype="int64")
        if not self.flat:
            ir.shape = (self.nd, self.nd, self.nd)
        return ir

    @property
    def fwidth(self):
        fw = np.ones((self.nd ** 3, 3), dtype="float64") / self.nd
        if not self.flat:
            fw.shape = (self.nd, self.nd, self.nd, 3)
        return self.ds.arr(fw, units="code_length")



# These are some of the data object registries that are used in different places in the
# code. Not all of the self-registering objects are included in these.

analysis_task_registry = {}
data_object_registry = {}
derived_quantity_registry = {}
output_type_registry = {}
simulation_time_series_registry = {}



"""
Generic file-handing functions for SPH data




"""
from yt.utilities.io_handler import BaseIOHandler


class IOHandlerSPH(BaseIOHandler):
    """IOHandler implementation specifically for SPH data

    This exists to handle particles with smoothing lengths, which require us
    to read in smoothing lengths along with the the particle coordinates to
    determine particle extents.
    """

    def _count_particles_chunks(self, psize, chunks, ptf, selector):
        if getattr(selector, "is_all_data", False):
            chunks = list(chunks)
            data_files = set()
            for chunk in chunks:
                for obj in chunk.objs:
                    data_files.update(obj.data_files)
            data_files = sorted(data_files, key=lambda x: (x.filename, x.start))
            for data_file in data_files:
                for ptype in ptf.keys():
                    psize[ptype] += data_file.total_particles[ptype]
        else:
            for ptype, (x, y, z), hsml in self._read_particle_coords(chunks, ptf):
                psize[ptype] += selector.count_points(x, y, z, hsml)
        return dict(psize)



from functools import wraps

import numpy as np

filter_registry = {}


def apply_filter(f):
    @wraps(f)
    def newfunc(*args, **kwargs):
        args[0]._filters.append((f.__name__, (args, kwargs)))
        return args[0]

    return newfunc


class FixedResolutionBufferFilter:

    """
    This object allows to apply data transformation directly to
    :class:`yt.visualization.fixed_resolution.FixedResolutionBuffer`
    """

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        filter_registry[cls.__name__] = cls

    def __init__(self, *args, **kwargs):
        pass

    def apply(self, buff):
        pass


class FixedResolutionBufferGaussBeamFilter(FixedResolutionBufferFilter):

    """
    This filter convolves
    :class:`yt.visualization.fixed_resolution.FixedResolutionBuffer` with
    2d gaussian that is 'nbeam' pixels wide and has standard deviation
    'sigma'.
    """

    _filter_name = "gauss_beam"

    def __init__(self, nbeam=30, sigma=2.0):
        self.nbeam = nbeam
        self.sigma = sigma

    def apply(self, buff):
        from yt.utilities.on_demand_imports import _scipy

        hnbeam = self.nbeam // 2
        sigma = self.sigma

        l = np.linspace(-hnbeam, hnbeam, num=self.nbeam + 1)
        x, y = np.meshgrid(l, l)
        g2d = (1.0 / (sigma * np.sqrt(2.0 * np.pi))) * np.exp(
            -((x / sigma) ** 2 + (y / sigma) ** 2) / (2 * sigma ** 2)
        )
        g2d /= g2d.max()

        npm, nqm = np.shape(buff)
        spl = _scipy.signal.convolve(buff, g2d)

        return spl[hnbeam : npm + hnbeam, hnbeam : nqm + hnbeam]


class FixedResolutionBufferWhiteNoiseFilter(FixedResolutionBufferFilter):

    """
    This filter adds white noise with the amplitude "bg_lvl" to
    :class:`yt.visualization.fixed_resolution.FixedResolutionBuffer`.
    If "bg_lvl" is not present, 10th percentile of the FRB's value is
    used instead.
    """

    _filter_name = "white_noise"

    def __init__(self, bg_lvl=None):
        self.bg_lvl = bg_lvl

    def apply(self, buff):
        if self.bg_lvl is None:
            amp = np.percentile(buff, 10)
        else:
            amp = self.bg_lvl
        npm, nqm = np.shape(buff)
        return buff + np.random.randn(npm, nqm) * amp



import numpy as np

from yt.utilities.io_handler import BaseIOHandler
from yt.utilities.logger import ytLogger as mylog


class IOHandlerFITS(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "fits"

    def __init__(self, ds):
        super().__init__(ds)
        self.ds = ds
        self._handle = ds._handle

    def _read_particles(
        self, fields_to_read, type, args, grid_list, count_list, conv_factors
    ):
        pass

    def _read_particle_coords(self, chunks, ptf):
        pdata = self.ds._handle[self.ds.first_image].data
        assert len(ptf) == 1
        ptype = list(ptf.keys())[0]
        x = np.asarray(pdata.field("X"), dtype="=f8")
        y = np.asarray(pdata.field("Y"), dtype="=f8")
        z = np.ones(x.shape)
        x = (x - 0.5) / self.ds.reblock + 0.5
        y = (y - 0.5) / self.ds.reblock + 0.5
        yield ptype, (x, y, z)

    def _read_particle_fields(self, chunks, ptf, selector):
        pdata = self.ds._handle[self.ds.first_image].data
        assert len(ptf) == 1
        ptype = list(ptf.keys())[0]
        field_list = ptf[ptype]
        x = np.asarray(pdata.field("X"), dtype="=f8")
        y = np.asarray(pdata.field("Y"), dtype="=f8")
        z = np.ones(x.shape)
        x = (x - 0.5) / self.ds.reblock + 0.5
        y = (y - 0.5) / self.ds.reblock + 0.5
        mask = selector.select_points(x, y, z, 0.0)
        if mask is None:
            return
        for field in field_list:
            fd = field.split("_")[-1]
            data = pdata.field(fd.upper())
            if fd in ["x", "y"]:
                data = (data.copy() - 0.5) / self.ds.reblock + 0.5
            yield (ptype, field), data[mask]

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        if any((ftype != "fits" for ftype, fname in fields)):
            raise NotImplementedError
        rv = {}
        dt = "float64"
        for field in fields:
            rv[field] = np.empty(size, dtype=dt)
        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s grids",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )
        dx = self.ds.domain_width / self.ds.domain_dimensions
        for field in fields:
            ftype, fname = field
            f = self.ds.index._file_map[fname]
            ds = f[self.ds.index._ext_map[fname]]
            bzero, bscale = self.ds.index._scale_map[fname]
            ind = 0
            for chunk in chunks:
                for g in chunk.objs:
                    start = ((g.LeftEdge - self.ds.domain_left_edge) / dx).d.astype(
                        "int"
                    )
                    end = start + g.ActiveDimensions
                    slices = [slice(start[i], end[i]) for i in range(3)]
                    if self.ds.dimensionality == 2:
                        nx, ny = g.ActiveDimensions[:2]
                        nz = 1
                        data = np.zeros((nx, ny, nz))
                        data[:, :, 0] = ds.data[slices[1], slices[0]].T
                    elif self.ds.naxis == 4:
                        idx = self.ds.index._axis_map[fname]
                        data = ds.data[idx, slices[2], slices[1], slices[0]].T
                    else:
                        data = ds.data[slices[2], slices[1], slices[0]].T
                    if fname in self.ds.nan_mask:
                        data[np.isnan(data)] = self.ds.nan_mask[fname]
                    elif "all" in self.ds.nan_mask:
                        data[np.isnan(data)] = self.ds.nan_mask["all"]
                    data = bzero + bscale * data
                    ind += g.select(selector, data.astype("float64"), rv[field], ind)
        return rv



import os
import shutil
import tempfile
import unittest

import numpy as np

import yt
from yt.data_objects.particle_filters import add_particle_filter
from yt.data_objects.profiles import Profile1D, Profile2D, Profile3D, create_profile
from yt.testing import (
    assert_equal,
    assert_raises,
    assert_rel_equal,
    fake_random_ds,
    fake_sph_orientation_ds,
    requires_module,
)
from yt.utilities.exceptions import YTIllDefinedProfile, YTProfileDataShape
from yt.visualization.profile_plotter import PhasePlot, ProfilePlot

_fields = ("density", "temperature", "dinosaurs", "tribbles")
_units = ("g/cm**3", "K", "dyne", "erg")


def test_profiles():
    ds = fake_random_ds(64, nprocs=8, fields=_fields, units=_units)
    nv = ds.domain_dimensions.prod()
    dd = ds.all_data()
    rt, tt, dt = dd.quantities["TotalQuantity"](
        [("gas", "density"), ("gas", "temperature"), ("stream", "dinosaurs")]
    )

    e1, e2 = 0.9, 1.1
    for nb in [8, 16, 32, 64]:
        for input_units in ["mks", "cgs"]:
            (rmi, rma), (tmi, tma), (dmi, dma) = (
                getattr(ex, f"in_{input_units}")()
                for ex in dd.quantities["Extrema"](
                    [
                        ("gas", "density"),
                        ("gas", "temperature"),
                        ("stream", "dinosaurs"),
                    ]
                )
            )
            # We log all the fields or don't log 'em all.  No need to do them
            # individually.
            for lf in [True, False]:
                direct_profile = Profile1D(
                    dd,
                    ("gas", "density"),
                    nb,
                    rmi * e1,
                    rma * e2,
                    lf,
                    weight_field=None,
                )
                direct_profile.add_fields([("index", "ones"), ("gas", "temperature")])

                indirect_profile_s = create_profile(
                    dd,
                    ("gas", "density"),
                    [("index", "ones"), ("gas", "temperature")],
                    n_bins=nb,
                    extrema={("gas", "density"): (rmi * e1, rma * e2)},
                    logs={("gas", "density"): lf},
                    weight_field=None,
                )

                indirect_profile_t = create_profile(
                    dd,
                    ("gas", "density"),
                    [("index", "ones"), ("gas", "temperature")],
                    n_bins=nb,
                    extrema={("gas", "density"): (rmi * e1, rma * e2)},
                    logs={("gas", "density"): lf},
                    weight_field=None,
                )

                for p1d in [direct_profile, indirect_profile_s, indirect_profile_t]:
                    assert_equal(p1d["index", "ones"].sum(), nv)
                    assert_rel_equal(tt, p1d["gas", "temperature"].sum(), 7)

                p2d = Profile2D(
                    dd,
                    ("gas", "density"),
                    nb,
                    rmi * e1,
                    rma * e2,
                    lf,
                    ("gas", "temperature"),
                    nb,
                    tmi * e1,
                    tma * e2,
                    lf,
                    weight_field=None,
                )
                p2d.add_fields([("index", "ones"), ("gas", "temperature")])
                assert_equal(p2d["index", "ones"].sum(), nv)
                assert_rel_equal(tt, p2d["gas", "temperature"].sum(), 7)

                p3d = Profile3D(
                    dd,
                    ("gas", "density"),
                    nb,
                    rmi * e1,
                    rma * e2,
                    lf,
                    ("gas", "temperature"),
                    nb,
                    tmi * e1,
                    tma * e2,
                    lf,
                    ("stream", "dinosaurs"),
                    nb,
                    dmi * e1,
                    dma * e2,
                    lf,
                    weight_field=None,
                )
                p3d.add_fields([("index", "ones"), ("gas", "temperature")])
                assert_equal(p3d["index", "ones"].sum(), nv)
                assert_rel_equal(tt, p3d["gas", "temperature"].sum(), 7)

        p1d = Profile1D(dd, ("index", "x"), nb, 0.0, 1.0, False, weight_field=None)
        p1d.add_fields(("index", "ones"))
        av = nv / nb
        assert_equal(p1d["index", "ones"], np.ones(nb) * av)

        # We re-bin ones with a weight now
        p1d = Profile1D(
            dd, ("index", "x"), nb, 0.0, 1.0, False, weight_field=("gas", "temperature")
        )
        p1d.add_fields([("index", "ones")])
        assert_equal(p1d["index", "ones"], np.ones(nb))

        # Verify we can access "ones" after adding a new field
        # See issue 988
        p1d.add_fields([("gas", "density")])
        assert_equal(p1d["index", "ones"], np.ones(nb))

        p2d = Profile2D(
            dd,
            ("index", "x"),
            nb,
            0.0,
            1.0,
            False,
            ("index", "y"),
            nb,
            0.0,
            1.0,
            False,
            weight_field=None,
        )
        p2d.add_fields(("index", "ones"))
        av = nv / nb ** 2
        assert_equal(p2d["index", "ones"], np.ones((nb, nb)) * av)

        # We re-bin ones with a weight now
        p2d = Profile2D(
            dd,
            ("index", "x"),
            nb,
            0.0,
            1.0,
            False,
            ("index", "y"),
            nb,
            0.0,
            1.0,
            False,
            weight_field=("gas", "temperature"),
        )
        p2d.add_fields([("index", "ones")])
        assert_equal(p2d["index", "ones"], np.ones((nb, nb)))

        p3d = Profile3D(
            dd,
            ("index", "x"),
            nb,
            0.0,
            1.0,
            False,
            ("index", "y"),
            nb,
            0.0,
            1.0,
            False,
            ("index", "z"),
            nb,
            0.0,
            1.0,
            False,
            weight_field=None,
        )
        p3d.add_fields(("index", "ones"))
        av = nv / nb ** 3
        assert_equal(p3d["index", "ones"], np.ones((nb, nb, nb)) * av)

        # We re-bin ones with a weight now
        p3d = Profile3D(
            dd,
            ("index", "x"),
            nb,
            0.0,
            1.0,
            False,
            ("index", "y"),
            nb,
            0.0,
            1.0,
            False,
            ("index", "z"),
            nb,
            0.0,
            1.0,
            False,
            weight_field=("gas", "temperature"),
        )
        p3d.add_fields([("index", "ones")])
        assert_equal(p3d["index", "ones"], np.ones((nb, nb, nb)))

        p2d = create_profile(
            dd,
            ("gas", "density"),
            ("gas", "temperature"),
            weight_field=("gas", "cell_mass"),
            extrema={("gas", "density"): (None, rma * e2)},
        )
        assert_equal(p2d.x_bins[0], rmi - np.spacing(rmi))
        assert_equal(p2d.x_bins[-1], rma * e2)
        assert str(ds.field_info["gas", "cell_mass"].units) == str(p2d.weight.units)

        p2d = create_profile(
            dd,
            ("gas", "density"),
            ("gas", "temperature"),
            weight_field=("gas", "cell_mass"),
            extrema={("gas", "density"): (rmi * e2, None)},
        )
        assert_equal(p2d.x_bins[0], rmi * e2)
        assert_equal(p2d.x_bins[-1], rma + np.spacing(rma))


extrema_s = {("all", "particle_position_x"): (0, 1)}
logs_s = {("all", "particle_position_x"): False}

extrema_t = {("all", "particle_position_x"): (0, 1)}
logs_t = {("all", "particle_position_x"): False}


def test_particle_profiles():
    for nproc in [1, 2, 4, 8]:
        ds = fake_random_ds(32, nprocs=nproc, particles=32 ** 3)
        dd = ds.all_data()

        p1d = Profile1D(
            dd, ("all", "particle_position_x"), 128, 0.0, 1.0, False, weight_field=None
        )
        p1d.add_fields([("all", "particle_ones")])
        assert_equal(p1d[("all", "particle_ones")].sum(), 32 ** 3)

        p1d = create_profile(
            dd,
            [("all", "particle_position_x")],
            [("all", "particle_ones")],
            weight_field=None,
            n_bins=128,
            extrema=extrema_s,
            logs=logs_s,
        )
        assert_equal(p1d[("all", "particle_ones")].sum(), 32 ** 3)

        p1d = create_profile(
            dd,
            [("all", "particle_position_x")],
            [("all", "particle_ones")],
            weight_field=None,
            n_bins=128,
            extrema=extrema_t,
            logs=logs_t,
        )
        assert_equal(p1d[("all", "particle_ones")].sum(), 32 ** 3)

        p2d = Profile2D(
            dd,
            ("all", "particle_position_x"),
            128,
            0.0,
            1.0,
            False,
            ("all", "particle_position_y"),
            128,
            0.0,
            1.0,
            False,
            weight_field=None,
        )
        p2d.add_fields([("all", "particle_ones")])
        assert_equal(p2d[("all", "particle_ones")].sum(), 32 ** 3)

        p3d = Profile3D(
            dd,
            ("all", "particle_position_x"),
            128,
            0.0,
            1.0,
            False,
            ("all", "particle_position_y"),
            128,
            0.0,
            1.0,
            False,
            ("all", "particle_position_z"),
            128,
            0.0,
            1.0,
            False,
            weight_field=None,
        )
        p3d.add_fields([("all", "particle_ones")])
        assert_equal(p3d[("all", "particle_ones")].sum(), 32 ** 3)


def test_mixed_particle_mesh_profiles():
    ds = fake_random_ds(32, particles=10)
    ad = ds.all_data()
    assert_raises(
        YTIllDefinedProfile,
        ProfilePlot,
        ad,
        ("index", "radius"),
        ("all", "particle_mass"),
    )
    assert_raises(
        YTIllDefinedProfile,
        ProfilePlot,
        ad,
        "radius",
        [("all", "particle_mass"), ("all", "particle_ones")],
    )
    assert_raises(
        YTIllDefinedProfile,
        ProfilePlot,
        ad,
        ("index", "radius"),
        [("all", "particle_mass"), ("index", "ones")],
    )
    assert_raises(
        YTIllDefinedProfile,
        ProfilePlot,
        ad,
        ("all", "particle_radius"),
        ("all", "particle_mass"),
        ("gas", "cell_mass"),
    )
    assert_raises(
        YTIllDefinedProfile,
        ProfilePlot,
        ad,
        ("index", "radius"),
        ("gas", "cell_mass"),
        ("all", "particle_ones"),
    )

    assert_raises(
        YTIllDefinedProfile,
        PhasePlot,
        ad,
        ("index", "radius"),
        ("all", "particle_mass"),
        ("gas", "velocity_x"),
    )
    assert_raises(
        YTIllDefinedProfile,
        PhasePlot,
        ad,
        ("all", "particle_radius"),
        ("all", "particle_mass"),
        ("gas", "cell_mass"),
    )
    assert_raises(
        YTIllDefinedProfile,
        PhasePlot,
        ad,
        ("index", "radius"),
        ("gas", "cell_mass"),
        ("all", "particle_ones"),
    )
    assert_raises(
        YTIllDefinedProfile,
        PhasePlot,
        ad,
        ("all", "particle_radius"),
        ("all", "particle_mass"),
        ("all", "particle_ones"),
    )


def test_particle_profile_negative_field():
    # see Issue #1340
    n_particles = int(1e4)

    ppx, ppy, ppz = np.random.normal(size=[3, n_particles])
    pvx, pvy, pvz = -np.ones((3, n_particles))

    data = {
        "particle_position_x": ppx,
        "particle_position_y": ppy,
        "particle_position_z": ppz,
        "particle_velocity_x": pvx,
        "particle_velocity_y": pvy,
        "particle_velocity_z": pvz,
    }

    bbox = 1.1 * np.array(
        [[min(ppx), max(ppx)], [min(ppy), max(ppy)], [min(ppz), max(ppz)]]
    )
    ds = yt.load_particles(data, bbox=bbox)
    ad = ds.all_data()

    profile = yt.create_profile(
        ad,
        [("all", "particle_position_x"), ("all", "particle_position_y")],
        ("all", "particle_velocity_x"),
        logs={
            ("all", "particle_position_x"): True,
            ("all", "particle_position_y"): True,
            ("all", "particle_position_z"): True,
        },
        weight_field=None,
    )
    assert profile[("all", "particle_velocity_x")].min() < 0
    assert profile.x_bins.min() > 0
    assert profile.y_bins.min() > 0

    profile = yt.create_profile(
        ad,
        [("all", "particle_position_x"), ("all", "particle_position_y")],
        ("all", "particle_velocity_x"),
        weight_field=None,
    )
    assert profile[("all", "particle_velocity_x")].min() < 0
    assert profile.x_bins.min() < 0
    assert profile.y_bins.min() < 0

    # can't use CIC deposition with log-scaled bin fields
    with assert_raises(RuntimeError):
        yt.create_profile(
            ad,
            [("all", "particle_position_x"), ("all", "particle_position_y")],
            ("all", "particle_velocity_x"),
            logs={
                ("all", "particle_position_x"): True,
                ("all", "particle_position_y"): False,
                ("all", "particle_position_z"): False,
            },
            weight_field=None,
            deposition="cic",
        )

    # can't use CIC deposition with accumulation or fractional
    with assert_raises(RuntimeError):
        yt.create_profile(
            ad,
            [("all", "particle_position_x"), ("all", "particle_position_y")],
            ("all", "particle_velocity_x"),
            logs={
                ("all", "particle_position_x"): False,
                ("all", "particle_position_y"): False,
                ("all", "particle_position_z"): False,
            },
            weight_field=None,
            deposition="cic",
            accumulation=True,
            fractional=True,
        )


def test_profile_zero_weight():
    def DMparticles(pfilter, data):
        filter = data[(pfilter.filtered_type, "particle_type")] == 1
        return filter

    def DM_in_cell_mass(field, data):
        return data["deposit", "DM_density"] * data["index", "cell_volume"]

    add_particle_filter(
        "DM", function=DMparticles, filtered_type="io", requires=["particle_type"]
    )

    _fields = (
        "particle_position_x",
        "particle_position_y",
        "particle_position_z",
        "particle_mass",
        "particle_velocity_x",
        "particle_velocity_y",
        "particle_velocity_z",
        "particle_type",
    )
    _units = ("cm", "cm", "cm", "g", "cm/s", "cm/s", "cm/s", "dimensionless")
    ds = fake_random_ds(
        32, particle_fields=_fields, particle_field_units=_units, particles=16
    )

    ds.add_particle_filter("DM")
    ds.add_field(
        ("gas", "DM_cell_mass"),
        units="g",
        function=DM_in_cell_mass,
        sampling_type="cell",
    )

    sp = ds.sphere(ds.domain_center, (10, "kpc"))

    profile = yt.create_profile(
        sp,
        [("gas", "density")],
        [("gas", "radial_velocity")],
        weight_field=("gas", "DM_cell_mass"),
    )

    assert not np.any(np.isnan(profile["gas", "radial_velocity"]))


def test_profile_sph_data():
    ds = fake_sph_orientation_ds()
    # test we create a profile without raising YTIllDefinedProfile
    yt.create_profile(
        ds.all_data(),
        [("gas", "density"), ("gas", "temperature")],
        [("gas", "kinetic_energy_density")],
        weight_field=None,
    )


def test_profile_override_limits():
    ds = fake_random_ds(64, nprocs=8, fields=_fields, units=_units)

    sp = ds.sphere(ds.domain_center, (10, "kpc"))
    obins = np.linspace(-5, 5, 10)
    profile = yt.create_profile(
        sp,
        [("gas", "density")],
        [("gas", "temperature")],
        override_bins={("gas", "density"): (obins, "g/cm**3")},
    )
    assert_equal(ds.arr(obins, "g/cm**3"), profile.x_bins)

    profile = yt.create_profile(
        sp,
        [("gas", "density"), ("stream", "dinosaurs")],
        [("gas", "temperature")],
        override_bins={
            ("gas", "density"): (obins, "g/cm**3"),
            ("stream", "dinosaurs"): obins,
        },
    )
    assert_equal(ds.arr(obins, "g/cm**3"), profile.x_bins)
    assert_equal(ds.arr(obins, "dyne"), profile.y_bins)

    profile = yt.create_profile(
        sp,
        [("gas", "density"), (("stream", "dinosaurs")), ("stream", "tribbles")],
        [("gas", "temperature")],
        override_bins={
            ("gas", "density"): (obins, "g/cm**3"),
            (("stream", "dinosaurs")): obins,
            ("stream", "tribbles"): (obins, "erg"),
        },
    )
    assert_equal(ds.arr(obins, "g/cm**3"), profile.x_bins)
    assert_equal(ds.arr(obins, "dyne"), profile.y_bins)
    assert_equal(ds.arr(obins, "erg"), profile.z_bins)


class TestBadProfiles(unittest.TestCase):

    tmpdir = None
    curdir = None

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()
        self.curdir = os.getcwd()
        os.chdir(self.tmpdir)

    def tearDown(self):
        os.chdir(self.curdir)
        # clean up
        shutil.rmtree(self.tmpdir)

    @requires_module("h5py")
    def test_unequal_data_shape_profile(self):
        density = np.random.random(128)
        temperature = np.random.random(128)
        mass = np.random.random((128, 128))

        my_data = {
            ("gas", "density"): density,
            ("gas", "temperature"): temperature,
            ("gas", "mass"): mass,
        }
        fake_ds_med = {"current_time": yt.YTQuantity(10, "Myr")}
        field_types = {field: "gas" for field in my_data.keys()}
        yt.save_as_dataset(fake_ds_med, "mydata.h5", my_data, field_types=field_types)

        ds = yt.load("mydata.h5")

        with assert_raises(YTProfileDataShape):
            yt.PhasePlot(
                ds.data,
                ("gas", "temperature"),
                ("gas", "density"),
                ("gas", "mass"),
            )

    @requires_module("h5py")
    def test_unequal_bin_field_profile(self):
        density = np.random.random(128)
        temperature = np.random.random(127)
        mass = np.random.random((128, 128))

        my_data = {
            ("gas", "density"): density,
            ("gas", "temperature"): temperature,
            ("gas", "mass"): mass,
        }
        fake_ds_med = {"current_time": yt.YTQuantity(10, "Myr")}
        field_types = {field: "gas" for field in my_data.keys()}
        yt.save_as_dataset(fake_ds_med, "mydata.h5", my_data, field_types=field_types)

        ds = yt.load("mydata.h5")

        with assert_raises(YTProfileDataShape):
            yt.PhasePlot(
                ds.data,
                ("gas", "temperature"),
                ("gas", "density"),
                ("gas", "mass"),
            )


def test_index_field_units():
    # see #1849
    ds = fake_random_ds(16, length_unit=2)
    ad = ds.all_data()
    icv_units = ad["index", "cell_volume"].units
    assert str(icv_units) == "code_length**3"
    gcv_units = ad["gas", "cell_volume"].units
    assert str(gcv_units) == "cm**3"
    prof = ad.profile(
        [("gas", "density"), ("gas", "velocity_x")],
        [("gas", "cell_volume"), ("index", "cell_volume")],
        weight_field=None,
    )
    assert str(prof["index", "cell_volume"].units) == "code_length**3"
    assert str(prof["gas", "cell_volume"].units) == "cm**3"


@requires_module("astropy")
def test_export_astropy():
    from yt.units.yt_array import YTArray

    ds = fake_random_ds(64)
    ad = ds.all_data()
    prof = ad.profile(
        ("index", "radius"),
        [("gas", "density"), ("gas", "velocity_x")],
        weight_field=("index", "ones"),
        n_bins=32,
    )
    # export to AstroPy table
    at1 = prof.to_astropy_table()
    assert "radius" in at1.colnames
    assert "density" in at1.colnames
    assert "velocity_x" in at1.colnames
    assert_equal(prof.x.d, at1["radius"].value)
    assert_equal(prof[("gas", "density")].d, at1["density"].value)
    assert_equal(prof[("gas", "velocity_x")].d, at1["velocity_x"].value)
    assert prof.x.units == YTArray.from_astropy(at1["radius"]).units
    assert prof[("gas", "density")].units == YTArray.from_astropy(at1["density"]).units
    assert (
        prof[("gas", "velocity_x")].units
        == YTArray.from_astropy(at1["velocity_x"]).units
    )
    assert np.all(at1.mask["density"] == prof.used)
    at2 = prof.to_astropy_table(fields=("gas", "density"), only_used=True)
    assert "radius" in at2.colnames
    assert "velocity_x" not in at2.colnames
    assert_equal(prof.x.d[prof.used], at2["radius"].value)
    assert_equal(prof[("gas", "density")].d[prof.used], at2["density"].value)


@requires_module("pandas")
def test_export_pandas():
    ds = fake_random_ds(64)
    ad = ds.all_data()
    prof = ad.profile(
        "radius",
        [("gas", "density"), ("gas", "velocity_x")],
        weight_field=("index", "ones"),
        n_bins=32,
    )
    # export to pandas DataFrame
    df1 = prof.to_dataframe()
    assert "radius" in df1.columns
    assert "density" in df1.columns
    assert "velocity_x" in df1.columns
    assert_equal(prof.x.d, df1["radius"])
    assert_equal(prof[("gas", "density")].d, np.nan_to_num(df1["density"]))
    assert_equal(prof["velocity_x"].d, np.nan_to_num(df1["velocity_x"]))
    df2 = prof.to_dataframe(fields=("gas", "density"), only_used=True)
    assert "radius" in df2.columns
    assert "velocity_x" not in df2.columns
    assert_equal(prof.x.d[prof.used], df2["radius"])
    assert_equal(prof[("gas", "density")].d[prof.used], df2["density"])
